"""BioAT package.
BioAT can be a package to import.
BioAT also can be a command-line tool.
It is a bioinformatic tool/pkg bundle for python.
"""

from bioat.bamtools import BamTools
from bioat.bedtools import BedTools
from bioat.crisprtools import CrisprTools
from bioat.exceptions import *  # !must be the first
from bioat.fastxtools import FastxTools
from bioat.hictools import HiCTools
from bioat.logger import __PKG_NAME__  # !must be the second
from bioat.metatools import MetaTools
from bioat.searchtools import SearchTools
from bioat.systemtools import SystemTools
from bioat.tabletools import TableTools
from bioat.target_seq import TargetSeq
from bioat.version import __author__, __doc_format__, __version__



"""Library for alignment, depends on Biopython

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

Library for alignment, depends on Biopython

example 1:
    bioat.lib.libalignment
        <in python consolo>:
            >>> from bioat.lib.libalignment import instantiate_pairwise_aligner
            >>> aligner = instantiate_pairwise_aligner()
            >>> alignment = get_best_alignment(
            >>>     Seq("GGCACTGCGGCTGGAAAAAAAAAAAAAAAGT"),
            >>>     Seq("GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT"),
            >>>     aligner=aligner,
            >>>     consider_strand=True
            >>> )
            >>> res = get_aligned_seq(alignment, reverse=False)
            >>> print(res)
            >>> res = get_aligned_seq(alignment, reverse=True)
            >>> print(res)
            >>> res = get_alignment_info(alignment, reverse=False)
            >>> print(res)
            >>> res = get_alignment_info(alignment, reverse=True)
            >>> print(res)
"""

import sys

from Bio.Align import Alignment, PairwiseAligner, substitution_matrices
from Bio.Seq import Seq

from bioat.logger import get_logger

__module_name__ = "bioat.lib.libalignment"


def instantiate_pairwise_aligner(
    scoring_match: float = 1,
    penalty_mismatch: float = 0.8,
    penalty_gap_open: float = -5,
    penalty_gap_extension: float = -2,
    penalty_query_left_gap_score: float = 0,
    penalty_query_right_gap_score: float = 0,
    mode="global",
    letn_match=False,
    score_matrix_dict: None | dict = None,
    log_level="DEBUG",
) -> PairwiseAligner:
    """Returns a PairwiseAligner object

    :param scoring_match: scoring_match, defaults to 1
    :type scoring_match: float, optional
    :param penalty_mismatch: penalty_mismatch, defaults to 0.8
    :type penalty_mismatch: float, optional
    :param penalty_gap_open: penalty_gap_open, defaults to -5
    :type penalty_gap_open: float, optional
    :param penalty_gap_extension: penalty_gap_extension, defaults to -2
    :type penalty_gap_extension: float, optional
    :param penalty_query_left_gap_score: penalty_query_left_gap_score, defaults to 0
    :type penalty_query_left_gap_score: float, optional
    :param penalty_query_right_gap_score: penalty_query_right_gap_score, defaults to 0
    :type penalty_query_right_gap_score: float, optional
    :param mode: global or local, defaults to "global"
    :type mode: str, optional
    :param letn_match: let base N as match or not, defaults to False
    :type letn_match: bool, optional
    :param log_level: log_level, defaults to "DEBUG"
    :type log_level: str, optional
    :param score_matrix_dict: DIY scoring matrix and other NT or AA, defaults to None
    :type score_matrix_dict: dict, optional
    :return: an object instantiated from PairwiseAligner
    :rtype: PairwiseAligner

    details for score_matrix_dict:
        you can use any base and any score, such as AGCTN, or 20AA + 2abnormalAA
        or ?:(), whatever character you want.
            score_matrix_dict = {
                ("A", "A"): scoring_match,
                ("A", "G"): penalty_mismatch,
                ("A", "C"): penalty_mismatch,
                ("A", "T"): penalty_mismatch,
                ("A", "N"): scoring_match,
                ("G", "A"): penalty_mismatch,
                ("G", "G"): scoring_match,
                ("G", "C"): penalty_mismatch,
                ("G", "T"): penalty_mismatch,
                ("G", "N"): scoring_match,
                ("C", "A"): penalty_mismatch,
                ("C", "G"): penalty_mismatch,
                ("C", "C"): scoring_match,
                ("C", "T"): penalty_mismatch,
                ("C", "N"): scoring_match,
                ("T", "A"): penalty_mismatch,
                ("T", "G"): penalty_mismatch,
                ("T", "C"): penalty_mismatch,
                ("T", "T"): scoring_match,
                ("T", "N"): scoring_match,
                ("N", "A"): scoring_match,
                ("N", "G"): scoring_match,
                ("N", "C"): scoring_match,
                ("N", "T"): scoring_match,
                ("N", "N"): scoring_match,
            }
    """
    logger = get_logger(
        level=log_level,
        module_name=__module_name__,
        func_name="instantiate_pairwise_aligner",
    )
    aligner = PairwiseAligner()

    if not letn_match and score_matrix_dict is None:
        # !如果 letn_match为False，
        # !并且未指定 score_matrix_dict
        # !则考虑本流程
        # !否则进入else中的自定义流程
        aligner.mode = mode
        aligner.match = scoring_match
        aligner.mismatch = penalty_mismatch
        aligner.open_gap_score = penalty_gap_open
        aligner.extend_gap_score = penalty_gap_extension
        aligner.query_left_gap_score = penalty_query_left_gap_score
        aligner.query_right_gap_score = penalty_query_right_gap_score
        aligner.wildcard = None
    else:
        if not score_matrix_dict:
            score_matrix_dict = {
                ("A", "A"): scoring_match,
                ("A", "G"): penalty_mismatch,
                ("A", "C"): penalty_mismatch,
                ("A", "T"): penalty_mismatch,
                ("A", "N"): scoring_match,
                ("G", "A"): penalty_mismatch,
                ("G", "G"): scoring_match,
                ("G", "C"): penalty_mismatch,
                ("G", "T"): penalty_mismatch,
                ("G", "N"): scoring_match,
                ("C", "A"): penalty_mismatch,
                ("C", "G"): penalty_mismatch,
                ("C", "C"): scoring_match,
                ("C", "T"): penalty_mismatch,
                ("C", "N"): scoring_match,
                ("T", "A"): penalty_mismatch,
                ("T", "G"): penalty_mismatch,
                ("T", "C"): penalty_mismatch,
                ("T", "T"): scoring_match,
                ("T", "N"): scoring_match,
                ("N", "A"): scoring_match,
                ("N", "G"): scoring_match,
                ("N", "C"): scoring_match,
                ("N", "T"): scoring_match,
                ("N", "N"): scoring_match,
            }

        substitution_matrix = substitution_matrices.Array(
            data=score_matrix_dict
        )
        aligner.mode = mode
        aligner.substitution_matrix = substitution_matrix
        # !自定义矩阵时，不可定义match和mismatch的打分，不然会出问题，在矩阵中给分即可
        # !aligner.match = scoring_match
        # !aligner.mismatch = penalty_mismatch
        aligner.open_gap_score = penalty_gap_open
        aligner.extend_gap_score = penalty_gap_extension
        aligner.query_left_gap_score = penalty_query_left_gap_score
        aligner.query_right_gap_score = penalty_query_right_gap_score
        aligner.wildcard = None
        # !自定义的打分矩阵中如果有"N"，则wildcard最好注释掉
        # !wildcard是一个特殊字符，用于在序列比对过程中表示一个可以与任何字符匹配的通配符
        # !设置wildcard的目的通常是为了在处理包含未知或可变字符的序列时提高比对的灵活性
        # !使用wildcard时需要谨慎，因为它可能会影响比对的准确性和生物学意义
        # !可以将wildcard设置为'N'，这样在比对过程中，字符'N'将被视为与任何字符都匹配

    logger.info(aligner)
    return aligner


def get_best_alignment(
    seq_a: Seq, seq_b: Seq, aligner: PairwiseAligner, consider_strand=True
):
    """Get best alignment by heightest alignment score

    Get best alignment by heightest alignment score,
    default to consider strand, that means, it will attempt to use
    seq_b.reverse_complement() to align

    :param seq_a: seq_a
    :type seq_a: Seq
    :param seq_b: seq_b
    :type seq_b: Seq
    :param aligner: aligner returns from instantiate_pairwise_aligner
    :type aligner: PairwiseAligner
    :param consider_strand: consider to use seq_b.reverse_complement() or not, defaults to True
    :type consider_strand: bool, optional
    :return: Alignment with a new attribute: alignment.is_a_reverse_complement_alignment (True / False)
    :rtype: Alignment
    """
    alignments = aligner.align(seq_a, seq_b)
    alignment = sorted(alignments, key=lambda x: x.score, reverse=True)[0]
    is_a_reverse_complement_alignment = None
    final_alignment = None

    if consider_strand:
        alignments_rc = aligner.align(seq_a, seq_b.reverse_complement())
        alignment_rc = sorted(
            alignments_rc, key=lambda x: x.score, reverse=True
        )[0]

        if alignment.score >= alignment_rc.score:
            final_alignment = alignment
            is_a_reverse_complement_alignment = False
        else:
            final_alignment = alignment_rc
            is_a_reverse_complement_alignment = True
    else:
        final_alignment = alignment
        is_a_reverse_complement_alignment = False

    # obj add attr
    final_alignment.is_a_reverse_complement_alignment = (
        is_a_reverse_complement_alignment
    )
    return final_alignment


def get_aligned_seq(
    alignment: Alignment, reverse: bool = False, letn_match=False
) -> dict:
    """Get_aligned_seq.

    :param alignment: Bio.Align.Alignment object
    :param reverse: return reversed seq info or not
    :return: dict, like this:
         {
            'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT',
            'aln_info':      '--..|.|||||||||||||||||||||||--||',
            'target_seq':    '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'
        }
    """
    ls = []

    for x in str(alignment).strip().split("\n"):
        if x:
            ls.append(x[20:].split(" ")[0])
    if not reverse:
        res = dict(
            reference_seq="".join(ls[::3]),
            aln_info="".join(ls[1::3]),
            target_seq="".join(ls[2::3]),
        )
    else:
        res = dict(
            reference_seq="".join(ls[::3])[::-1],
            aln_info="".join(ls[1::3])[::-1],
            target_seq="".join(ls[2::3])[::-1],
        )
    if letn_match:
        length = len(res["reference_seq"])
        # change base N to match
        i_n_a = [
            i
            for i in range(length)
            if (res["reference_seq"][i] == "N")
            and (res["target_seq"][i] != "-")
        ]
        i_n_b = [
            i
            for i in range(length)
            if (res["target_seq"][i] == "N")
            and (res["reference_seq"][i] != "-")
        ]
        i_n = list(set(i_n_a) | set(i_n_b))
        aln_list = list(res["aln_info"])

        for i in sorted(i_n):
            aln_list[i] = "★"
        res["aln_info"] = "".join(aln_list)

    return res


def get_alignment_info(
    alignment: Alignment,
    reverse: bool = False,
    letn_match=False,
    log_level="WARNING",
):
    """

    :param alignment: Bio.Align.Alignment object
    :param reverse: return reversed seq info or not
    :return:
            1. match count
            2. mismatch count
            3. gap count
            4. alignment.score
    :doc:
            The gap count should be the num of gap contain in sgRNA alignment region.

            e.g.

            AGTGGTAAGAAGAAGACGAGACATAATGAG
            ------||||||||||||||.|----||||
            ------AAGAAGAAGACGAGCC----TGAG

            gap count should be 4, rather than 10.

            add return info, start_index, end_index, now the retrun list will be

            return_list = [
                match_count,
                mismatch_count,
                gap_count,
                alignment.score,
                start_index,
                end_index
            ]

            The <start_index> and <end_index> are index related to sgRNA alignment string
    """
    # set logger
    logger = get_logger(
        level=log_level,
        module_name=__module_name__,
        func_name="get_alignment_info",
    )

    if reverse:
        aln_res = get_aligned_seq(
            alignment, reverse=True, letn_match=letn_match
        )
    else:
        aln_res = get_aligned_seq(alignment, letn_match=letn_match)

    reference_seq = aln_res["reference_seq"]
    aln_info = aln_res["aln_info"]
    target_seq = aln_res["target_seq"]
    aln_score = alignment.score
    target_seq_length = len(target_seq)
    target_seq_length_rm_gap = len(target_seq.replace("-", ""))
    aligned_coordinates = alignment.aligned[::-1]
    if hasattr(alignment, "is_a_reverse_complement_alignment"):
        is_a_reverse_complement_alignment = (
            alignment.is_a_reverse_complement_alignment
        )
    else:
        is_a_reverse_complement_alignment = None

    logger.debug(f"reference_seq = {reference_seq}")
    logger.debug(f"aln_info      = {aln_info}")
    logger.debug(f"target_seq    = {target_seq}")
    logger.debug(f"aln_score     = {aln_score}")
    logger.debug(f"target_seq_length (remove gap) = {target_seq_length_rm_gap}")
    logger.debug(
        f"is_a_reverse_complement_alignment = {is_a_reverse_complement_alignment}"
    )
    logger.debug(f"aligned_coordinates =\n{aligned_coordinates}")

    # define params
    ref_aln_start = target_seq_length - len(target_seq.lstrip("-"))
    ref_aln_end = ref_aln_start + len((target_seq.strip("-")))
    match_count = 0
    mismatch_count = 0
    gap_count = 0
    parsed_target_bases = 0
    encounter_target_seq = False

    for idx, base_info in enumerate(aln_info):
        """
        ---TTCTGCCTCTGGAGAGGG-GGAGGGGCCT---
        -------|.|.|||..|-.||-||.||--------
        -------GGCACTGCGG-TGGAGGTGG--------
        """
        if not encounter_target_seq:
            if target_seq[idx] == "-":
                # if start with -
                # not encounter with target_seq yet!
                # print(idx, base_info)
                continue
            else:
                # find first aligned base
                # encounter with target_seq!
                encounter_target_seq = True

        if encounter_target_seq and (
            parsed_target_bases < target_seq_length_rm_gap
        ):
            # from the same loop of [else: catch_start = True]
            # parse start base and later base
            # print(parsed_target_bases, target_seq_length_rm_gap, base_info)
            if base_info == "|" or base_info == "★":
                # match
                match_count += 1
                parsed_target_bases += 1
            elif base_info == ".":  # . instead of X after Biopython 1.8.0
                # mismatch
                mismatch_count += 1
                parsed_target_bases += 1
            elif base_info == "-":
                # gap on reference_seq or target_seq
                gap_count += 1

                if reference_seq[idx] == "-":
                    # gap on reference_seq
                    # value base on target_seq
                    parsed_target_bases += 1
                    pass
                else:
                    # gap on target_seq
                    # no value base on target_seq
                    parsed_target_bases += 0
                    pass
            else:
                logger.critical(
                    f"find not surpported alignment symbol: {base_info} ({idx} on the reference), check"
                    f"whether Biopython >= 1.8.0 or not"
                )
                exit(1)
        else:
            # parsed_target_bases = target_seq_length_rm_gap
            # full target_seq bases were processed
            break

    res = dict(
        match_count=match_count,
        mismatch_count=mismatch_count,
        gap_count=gap_count,
        aln_score=aln_score,
        ref_aln_start=ref_aln_start,
        ref_aln_end=ref_aln_end,
        is_a_reverse_complement_alignment=is_a_reverse_complement_alignment,
        alignment=aln_res,
    )
    return res



import datetime
import gzip
import json
import os
import re
import sys
import tarfile
import textwrap
import time
import xml.etree.ElementTree as ET
from collections import defaultdict
from hashlib import md5

import requests
from requests import HTTPError
from requests.exceptions import ChunkedEncodingError
from requests.utils import cookiejar_from_dict
from tqdm import tqdm
from urllib3.exceptions import InvalidChunkLength, ProtocolError

from bioat.exceptions import BioatParameterFormatError
from bioat.lib.libpath import HOME
from bioat.lib.libspider import ProxyPool, get_random_user_agents
from bioat.logger import get_logger

__module_name__ = "bioat.lib.libjgi"


class JGIDoc:
    DEFAULT_CATEGORIES = [
        "ESTs",
        "EST Clusters",
        "Assembled scaffolds (unmasked)",
        "Assembled scaffolds (masked)",
        "Transcripts",
        "Genes",
        "CDS",
        "Proteins",
        "Additional Files",
    ]
    usage_example_blurb = """\
        This tool will retrieve files from JGI
        It will return a list of possible files for downloading.
    
        To get <jgi_address>, go to: http://genome.jgi.doe.gov/ and search for your
        species of interest. Click through until you are at the "Info" page. For
        \x1B[3mNematostella vectensis\x1B[23m, the appropriate page is
        "http://genome.jgi.doe.gov/Nemve1/Nemve1.info.html".
    
        To query using only the name simply requires the specific JGI organism
        abbreviation, as referenced in the full url.
    
        For the above example, the proper input syntax for this script would be:
    
        $ bioat meta JGI_query -q http://genome.jgi.doe.gov/Nemve1/Nemve1.info.html
    
                                 -or-
    
        $ bioat meta JGI_query -q Nemve1
    
        If you already have the XML file for the query in the directory, you may use
        the --xml flag to avoid redownloading it (particularly useful if querying
        large, top-level groups with many sub-species, such as "fungi"):
    
        $ bioat meta JGI_query -x <your_xml_index>
        
        If the XML filename is omitted when using the -x/--xml flag, it is assumed that
        the XML file is named "jgi-xml-query.result_<organism-name>.xml". In such cases, the
        organism name is required.
        """
    select_blurb = """\
        # SYNTAX ///////////////////////////////////////////////////////////////////////
        Using the following format syntax to download selected file:
            <category number>:<i>[,<i>, <i>];<category number>:<i>-<i>;...
        
        Indices (<i>) may be a mixture of comma-separated values and hyphen-separated 
        ranges. 
        
        Example:
            '3:4,5; 7:1-10,13' will select elements 4 and 5 from category 3, and 1-10 
            plus 13 from category 7.
        # /SYNTAX ///////////////////////////////////////////////////////////////////////
        """


class JGIConfig:
    URL_JGI_MAIN = "https://genome.jgi.doe.gov"  # url home for jgi-img database
    URL_JGI_LOGIN = "https://signon.jgi.doe.gov/signon/create"  # url login
    URL_JGI_FETCH_XML = "https://genome.jgi.doe.gov/portal/ext-api/downloads/get-directory"  # url fetch xml
    FILENAME_TEMPLATE_XML = "jgi-xml-query.result_{}.xml"
    FILENAME_TEMPLATE_LOG_FAIL = "jgi-xml-query.failed_{}.log"
    FILENAME_COOKIE = os.path.join(HOME, ".bioat", "JGI", "cookie")
    FILENAME_CONFIG_PATH = os.path.join(HOME, ".bioat", "JGI", "account.conf")

    def __init__(self, overwrite_conf: bool = False):
        self.logger = get_logger(
            level="DEBUG", module_name=__module_name__, func_name="JGIConfig"
        )
        self.info: dict = {"user": None, "password": None, "categories": None}
        self.logger.debug("start to set / load user info...")
        if overwrite_conf:
            # 只要指定重写参数就重写，无需判断其他
            self.input_user_info()  # 从终端手动输入JGI用户信息
            self.save_config()  # 保存用户信息
            # 然后直接退出
            self.logger.info(
                "Configuration complete.\n"
                "Script may now be used to query JGI.\n"
                "You need re-run your command without parameter `--overwrite_conf`\n"
            )
            # self.load_config()  # 从文件加载用户信息
        else:
            if (
                    os.path.isfile(self.FILENAME_CONFIG_PATH)
                    and os.path.getsize(self.FILENAME_CONFIG_PATH) > 0
            ):
                # 直接加载配置信息
                self.load_config()
            else:
                # 配置信息保存文件异常
                self.logger.warning(
                    f"errors occur when loading file: {self.FILENAME_CONFIG_PATH}!\n"
                    f"try to get user info by manual inputting..."
                )
                self.input_user_info()  # 从终端手动输入JGI用户信息
                self.save_config()  # 保存用户信息
                self.load_config()  # 从文件加载用户信息
        self.logger.debug("set / load user info success!")

    def load_config(self):
        """
        Reads "user", "password" and "categories" entries
        from config file.

        """
        self.logger.debug("loading JGI account info...")

        with open(self.FILENAME_CONFIG_PATH, "rt") as f:
            for line in f:
                line = line.strip()
                if line.startswith("user"):
                    self.info["user"] = line.split("=")[1]
                if line.startswith("password"):
                    self.info["password"] = line.split("=")[1]
                if line.startswith("categories"):
                    cats = line.strip().split("=")[1]
                    self.info["categories"] = [e.strip() for e in cats.split(",")]

        if not all([self.info["user"], self.info["password"]]):
            self.logger.critical(
                f"Config file present ({self.FILENAME_CONFIG_PATH}), but user and/or password not found."
            )
            sys.exit(1)

        self.logger.debug("loading JGI account info done")

    def input_user_info(self):
        """
        Dialog with user to gather user information for
        use with the curl query. Returns a dict.

        """
        blurb = """
        === USER SETUP ===

        JGI access configuration:

        Before continuing, you will need to provide your JGI login credentials.
        These are required by JGI's curl api, and will be stored in a config
        file for future use (unless you choose to delete them).

        If you need to sign up for a JGI account, use the registration link at
        https://contacts.jgi.doe.gov/registration/new

        === CREDENTIALS ===
        """
        # Print left-justified triple-quoted text blocks
        print(textwrap.dedent(blurb))
        user_query = "JGI account username/email (or 'q' to quit): "

        # write user and passwd from interactive terminal
        self.info["user"] = input(user_query)

        if self.info["user"] == "q":
            sys.exit("Exiting now.")

        self.info["password"] = input("JGI account password (or 'q' to quit): ")

        if self.info["password"] == "q":
            sys.exit("Exiting now.")

        if self.info["user"].strip() == "" or self.info["password"] == "":
            self.logger.warning("user and password can not be None!")
            self.input_user_info()

        input_blurb = (
            f"Proceed with USER={self.info['user']}, PASSWORD={self.info['password']} to configure "
            "script?\n([y]es, [n]o, [r]estart): "
        )

        self.info["categories"]: list[str] = JGIDoc.DEFAULT_CATEGORIES

        while True:  # catch invalid responses
            choice = input(input_blurb)

            if choice.lower() == "y":
                break
            elif choice.lower() == "n":
                sys.exit("Exiting now.")
            elif choice.lower() == "r":
                self.input_user_info()

    def save_config(self):
        """
        Creates a config file <config_path> using
        credentials from dict <config_info>.

        """
        self.logger.debug("saving JGI account info...")
        u = self.info["user"]
        p = self.info["password"]
        c = self.info["categories"]
        c = ",".join(c)
        header = "# bioat meta JGI_query: JGI account info and essentials {}\n".format(
            "#" * 24
        )
        info = f"user={u}\npassword={p}\ncategories={c}"

        if not os.path.exists(os.path.dirname(self.FILENAME_CONFIG_PATH)):
            os.makedirs(os.path.dirname(self.FILENAME_CONFIG_PATH))

        with open(self.FILENAME_CONFIG_PATH, "wt") as f:
            f.write(header)
            f.write(info)

        self.logger.debug("saving JGI account info done")


class JGIOperator:
    def __init__(
            self,
            # pick one from three
            query_info: str | None = None,
            xml: str | None = None,
            log_fails: str | None = None,
            # runtime params
            nretry: int = 4,
            timeout: int = -1,
            regex: str | None = None,
            all_get: bool = False,
            overwrite_conf: bool = False,
            filter_files: bool = False,
            proxy_pool: str | None = None,
            just_query_xml: bool = False,
            # doc helper
            syntax_help: bool = False,
            usage: bool = False,
            # log
            log_level: str = "INFO",
    ):
        logger = get_logger(level=log_level, module_name=__module_name__, func_name="class: JGIOperator.__init__")
        # from cmd parameters
        self.query_info = query_info
        self.xml = xml
        self.log_fails = log_fails
        self.nretry = nretry
        self.timeout = None if timeout == -1 else timeout
        self.regex = regex
        self.all_get = all_get
        self.overwrite_conf = overwrite_conf
        self.filter_files = filter_files  # TODO 搞清楚用法
        self.just_query_xml = just_query_xml
        self.syntax_help = syntax_help
        self.usage = usage
        self.log_level = log_level
        # self
        self._dict_to_get = dict()
        self._urls_to_get = set()
        self._url_to_validate = defaultdict(dict)
        self._selections = dict()
        self._downloaded_files = list()
        self._failed_urls = list()
        self._log_fails_loaded = list()
        self._desired_categories = dict()
        self._cookie = None
        self._user_agent = get_random_user_agents()
        # From other obj #
        # load configs; auto check if you need overwrite user info or not
        self.config = JGIConfig(overwrite_conf=overwrite_conf)
        self.interactive = False
        # load docs;
        self.docs = JGIDoc
        # load proxy
        if proxy_pool:
            self._proxy_pool = ProxyPool(url=proxy_pool)
            self._proxy_ip = self._proxy_pool.get_proxy().get("proxy")
            self._proxies = {"http": f"http://{self._proxy_ip}"}
            logger.info(f"use proxy mode! proxy_pool = {proxy_pool}")
            logger.info(f"now proxy IP = {self._proxy_ip}")
        else:
            self._proxy_pool = None
            self._proxy_ip = None
            self._proxies = None
        # / From other obj
        logger.debug("checker for doc mode")
        self._run_doc()
        logger.debug("checker for input")
        self._parse_input()
        logger.debug("run login")
        self._load_cookie()

    # step 01 print and exit
    def _run_doc(self):
        """Checker for doc mode.

        print syntax_help and/or usage if these parameters are defined. And then, exit progress.
        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_run_doc",
        )
        run_docs = any([self.syntax_help, self.usage])
        if run_docs:
            logger.debug("doc mode is selected")
            # Check if user wants query help
            if self.syntax_help:
                print(f"\n[syntax_help]:\n{self.docs.select_blurb}")
            if self.usage:
                print(f"\n[usage]:\n{self.docs.usage_example_blurb}")
            sys.exit("Done. exit.")
            # finally exit

    # step 02 update self.query_info and self.log_fails and followed by self.login
    def _parse_input(self):
        """Checker for input.

        A checker for this rule: ONLY ONE of the parameters ('query_info', 'xml' and 'log_fails') can be specified;
        Parse query_info/xml/log_fails to update self.query_info and self.log_fails.
        After this method, you should call self.login method and then self.query method
        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_parse_input",
        )
        logger.debug(
            "update JGIOperator obj.query_info and obj.log_fails using parameters:"
            " 'query_info', 'xml' and 'log_fails'"
        )

        if sum(map(bool, [self.query_info, self.xml, self.log_fails])) != 1:
            # 这三个参数只可以指定一个！
            logger.error(
                "ONLY ONE of the parameters ('query_info', 'xml' and 'log_fails') can be specified!"
            )
            sys.exit("Done. exit.")
        else:
            logger.debug("pass param checker: 'query_info', 'xml' and 'log_fails'")

        if self.query_info:
            try:
                # if query_info is a URL
                logger.debug("attempt to parse query_info as a url")
                # query_regex = re.compile(r'\.jgi.+\.(?:gov|org).*\/(.+)\/(?!\/)')
                query_regex = re.compile(r"\.jgi.+\.(?:gov|org).*/(.+)/(?!/)")
                self.query_info = query_regex.search(self.query_info).group(1)
            except AttributeError:
                # if query_info is an organism name
                logger.debug(
                    "not a url, try to define query_info as an organism name abbreviation"
                )
                # query_info = query_info
        elif self.log_fails:
            # load failed info from log file if provided
            # filename see: class JGIConfig.FILENAME_TEMPLATE_LOG_FAIL
            self.query_info = (
                os.path.basename(self.log_fails).split(".")[1].replace("failed_", "")
            )

            logger.debug(
                f"get self.query_info = {self.query_info} from self.log_fails = {self.log_fails}"
            )
            with open(self.log_fails, "rt") as f:
                self._log_fails_loaded = f.read().splitlines()
            logger.debug(
                f"get self._log_fails_loaded = {self._log_fails_loaded} from self.log_fails = {self.log_fails}"
            )

        elif self.xml:
            # parse query_info from xml file content
            name_pattern = r"name=\"(.+)\""
            _org_line = None
            with open(self.xml, "rt") as f:
                for line in f:
                    if "organismDownloads" in line:
                        # standardized name indicator
                        # <organismDownloads name="Nemve1">
                        _org_line = line.strip()
                        break  # don't keep looking, already found
            try:
                self.query_info = re.search(name_pattern, _org_line).group(1)
            except TypeError:  # org_line still None
                logger.critical("the xml file seems wrong")
                sys.exit("Exit with errors.")
        else:
            logger.error(
                "one of the parameters 'query_info' and 'xml' should be specified"
            )
            sys.exit("Done. exit.")

    # step 03 login
    def _load_cookie(self):
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_load_cookie",
        )

        # prepare info
        url = self.config.URL_JGI_LOGIN
        cookie_file = self.config.FILENAME_COOKIE

        if os.path.exists(cookie_file):
            # if local
            exist_time = (
                    datetime.date.fromtimestamp(os.path.getmtime(cookie_file))
                    - datetime.date.today()
            )

            if exist_time <= datetime.timedelta(-1):
                logger.info("cookie is expired, try to re-loginning")
                os.remove(cookie_file)
                self._load_cookie()
            else:
                with open(cookie_file, "rt") as f:
                    cookies_dict = json.loads(f.read())
                self._cookie = requests.utils.cookiejar_from_dict(cookies_dict)
                logger.debug(
                    f"update self._cookie from [local], self._cookie = {self._cookie}"
                )
        else:
            logger.debug("update self._cookie from [JGI website], trying...")
            headers = {"User-Agent": self._user_agent}
            data = {
                "login": self.config.info["user"],
                "password": self.config.info["password"],
                "commit": "Sign In",
            }

            retry_count = 0
            while True:
                with requests.session() as s:  # set session
                    # requests.session和直接访问的主要区别在于会话的保持和请求效率。
                    # - 会话保持, 会话能让我们在跨请求的时候保持某些参数，比如在同一个session实例发出的所有请求之间保持cookie信息。
                    #            这对于需要登录状态的请求特别有用，因为登录信息可以在整个会话期间保持，而不需要每次请求时都重新输入。
                    # - 请求效率: 使用requests.session可以避免在每次发送请求时都将cookie信息放到请求内容中，因为session对象能够
                    #            自动获取到cookie并且可以在下一次请求时自动带上。这可以提高请求效率。
                    if retry_count >= self.nretry:
                        logger.error(
                            f"retry_count ({retry_count}) reaching max of self.nretry ({self.nretry})..."
                        )
                        self._clean_exit(  # exit
                            exit_message="exit with error",
                            exit_code=1,
                            rm_cookie=True,
                            rm_xml=False if self.xml else False,
                        )  # exit
                    try:
                        with s.post(
                                url,
                                headers=headers,
                                data=data,
                                timeout=self.timeout,
                                proxies=self._proxies,
                        ) as response:
                            logger.debug("request.post seems success")
                            # check response status code
                            logger.debug(
                                f"check response status code: {response.status_code}"
                            )
                            if response.status_code == 200:
                                logger.debug("login successes, get cookie...")
                                # save cookie to file
                                logger.debug(f"save cookie to {cookie_file}")
                                cookies_dict = requests.utils.dict_from_cookiejar(
                                    s.cookies
                                )
                                cookies_str = json.dumps(cookies_dict)

                                with open(cookie_file, "wt") as f:
                                    f.write(cookies_str)
                                logger.debug(
                                    f"successfully login, write cookie @ {cookie_file}"
                                )
                                break  # 跳出循环进入
                            else:
                                logger.critical(
                                    "Couldn't connect with server. Please check Internet connection and retry."
                                )
                                self._clean_exit(
                                    exit_message="exit with error",
                                    exit_code=1,
                                    rm_cookie=True,
                                    rm_xml=False if self.xml else False,
                                )
                    except Exception:
                        # 删除代理池中代理 if self._proxy_ip not None
                        logger.debug(
                            f"request.post seems fail, remove proxy ({self._proxy_ip}) in pool"
                        )
                        self._proxy_pool.delete_proxy(self._proxy_ip)
                        retry_count += 1

    # step 04 query
    def query(self):
        logger = get_logger(
            level=self.log_level, module_name=__module_name__, func_name="query"
        )
        # prepare info
        url = self.config.URL_JGI_FETCH_XML
        cookie_file = self.config.FILENAME_COOKIE
        headers = {"User-Agent": self._user_agent}
        params = {"organism": self.query_info}

        with open(cookie_file, "rt") as f:
            cookies_dict = json.loads(f.read())
        cookies = requests.utils.cookiejar_from_dict(cookies_dict)
        with requests.get(
                url,
                params=params,
                cookies=cookies,
                allow_redirects=True,
                stream=True,
                headers=headers,
                timeout=self.timeout,
                # proxies=self._proxies
        ) as response:
            logger.debug(f"login url requests.get: {response.url}")
            if self._proxies:
                logger.debug(f"query using IP: {self._proxy_ip}")
            try:
                # 如果响应的状态码不是200，将引发HTTPError异常
                response.raise_for_status()
            except HTTPError:
                logger.critical(f"response status: {response.status_code}")
                logger.critical(
                    "Couldn't connect with server. "
                    "Please check Internet connection (or accession rights) and retry."
                )
                logger.critical(f"response.text = {response.text}")
                self._clean_exit(
                    exit_message="exit with HTTPError",
                    exit_code=1,
                    rm_cookie=True,
                    rm_xml=True,
                )

            xml_file = self.config.FILENAME_TEMPLATE_XML.format(self.query_info)
            with open(xml_file, "wb") as f:
                # 使用二进制写入模式（"wb"）来保存结果文件，
                # 因为response.content返回的是一个字节字符串
                logger.debug(f"successfully query, write xml @ {xml_file}")
                f.write(response.content)

            # if just_query_xml = True, exit from here
            if self.just_query_xml:
                self._clean_exit(
                    exit_message="exit reason: just_query_xml = True",
                    exit_code=0,
                    rm_cookie=True,
                    rm_xml=False,
                )

    # step 05 parse xml info to update url
    def parse_xml(self):
        """
        Moves through the xml document <xml_file> and returns information
        about matches to elements in <DESIRED_CATEGORIES> if
        <filter_categories> is True, or all files otherwise

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="parse_xml",
        )

        # Parse xml file for content to download

        xml_file = self.config.FILENAME_TEMPLATE_XML.format(self.query_info)

        logger.debug(
            f"start to parse_xml using parameter filter_categories = {self.filter_files}"
        )
        display_cats = [
            "filename",
            "url",
            "size",
            "label",
            "sizeInBytes",
            "timestamp",
            "md5",
        ]
        # Choose between different XML parsers
        # Will only be used if --filter_files flag
        # if filter_files, user wants only those files in <_desired_categories>
        logger.debug("_xml_hunt...")
        found = self._xml_hunt(xml_file)
        found = self._format_found(found, self.filter_files)

        if not list(found.values()):
            return None
        logger.debug(f"successfully parsed xml @ {xml_file}")
        logger.debug("start to update file_list")

        category_id = 0

        for category, sub_cat in sorted(found.items()):
            if category not in self._desired_categories:
                category_id += 1
                self._desired_categories[category] = defaultdict(dict)
                self._desired_categories[category]["catID"] = category_id
            uid = 1
            for parent, children in sorted(sub_cat.items()):
                self._desired_categories[category]["results"][parent] = defaultdict(
                    dict
                )
                results = self._desired_categories[category]["results"][parent]
                unique_children = self._uniqueify(children)
                for child in sorted(unique_children, key=lambda x: x["filename"]):
                    try:
                        results[uid]
                    except KeyError:
                        results[uid] = {}
                    for dc in display_cats:
                        try:
                            results[uid][dc] = child[dc]
                        except KeyError:
                            continue
                    uid += 1

        logger.debug(
            f"successfully update self._desired_categories = {str(self._desired_categories)[:400]}...(omit)..."
        )

        # Check if file has any categories of interest
        if not any(v["results"] for v in list(self._desired_categories.values())):
            logger.error(
                "no results found for '{}' in any of the following "
                "categories:\n---\n{}\n---".format(
                    self.query_info, "\n".join(self.config.info["categories"])
                )
            )
            self._clean_exit(
                exit_message="exit with error",
                exit_code=1,
                rm_cookie=False,
                rm_xml=False if self.xml else False,
            )

    # step 06 download from url
    def download(self):
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="download",
        )
        self._decision_tree()

        self._urls_to_get = sorted(self._urls_to_get)
        filenames = [u.split("/")[-1] for u in self._urls_to_get]

        logger.debug(
            f"self._desired_categories = {str(self._desired_categories)[:400]}...(omit)..."
        )
        file_sizes = self._get_sizes(self._desired_categories, sizes_by_url={})
        # logger.debug(f'file_sizes = {file_sizes}')
        total_size = sum(filter(None, [file_sizes[url] for url in self._urls_to_get]))
        size_string = self._byte_convert(total_size)
        logger.info(
            f"Total download size for {len(self._urls_to_get)} files: {size_string}"
        )

        if self.interactive:
            select = input(
                "Continue? (y/n/b/p for yes/no/back/preview files): "
            ).lower()
            if select == "p":
                while select == "p":
                    print("\n".join(filenames))
                    select = input(
                        "Continue? (y/n/b/p for yes/no/back/preview files): "
                    ).lower()
            if select == "n":
                self._clean_exit(
                    exit_message="ABORTING DOWNLOAD",
                    exit_code=0,
                    rm_cookie=False,
                    rm_xml=False if self.xml else False,
                )
            elif select == "b":
                logger.info("back to select files...")
                self.interactive = True
                logger.debug("!!! re-call self.download position1")
                self.download()
            elif select == "y" or select == "":
                self._failed_urls, _ = self._download_list(self._urls_to_get)
            else:
                logger.info("illegal selection, back to select files...")
                self.interactive = True
                logger.debug("!!! re-call self.download position2")
                self.download()
        else:  # non interactive
            if self.regex or self.all_get or self.log_fails:
                self._failed_urls, _ = self._download_list(self._urls_to_get)
            else:
                raise BioatParameterFormatError

        logger.info(
            "Finished downloading {} files.".format(len(self._downloaded_files))
        )

        failed_happen = False
        if self.interactive:
            if self._failed_urls:
                n_broken = len(self._failed_urls)
                nretry_broken = input(
                    "{} files failed to download; nretry them? (y/n): ".format(n_broken)
                )
                if nretry_broken.lower() in ("yes", "y"):
                    self._failed_urls, _ = self._download_list(self._failed_urls)
            # Kindly offer to unpack files, if files remain after error check
            if self._downloaded_files:
                decompress = input(
                    (
                        "Decompress all downloaded files? "
                        "(y/n/k, k=decompress and keep original): "
                    )
                )
                if decompress != "n":
                    if decompress == "k":
                        keep_original = True
                    else:
                        keep_original = False
                    self._decompress_files(self._downloaded_files, keep_original)
                    logger.info("Finished decompressing all files.")
        else:
            # non-interactive
            if self._failed_urls:
                # Write failed URLs to a local log file.
                fail_log_file = self.config.FILENAME_TEMPLATE_LOG_FAIL.format(
                    self.query_info
                )
                logger.info(
                    f"{len(self._failed_urls)} failed downloads record into {fail_log_file}"
                )
                # write failed URLs to local file
                with open(fail_log_file, "wt") as f:
                    f.write("\n".join(self._failed_urls))
                failed_happen = True
            else:
                failed_happen = False

        # Clean up and exit

        if self.interactive:
            # interactive
            keep_temp = input(
                f"Keep temporary files ('{self.config.FILENAME_TEMPLATE_XML.format(self.query_info)}' "
                f"and '{self.config.FILENAME_COOKIE}')? (y/n): "
            )
            exit_message = "User choose to exit"
            if keep_temp.lower() in "y, yes":
                rm_cookie = False
                rm_xml = False
            else:
                rm_cookie = True
                rm_xml = True
        else:
            # non-interactive
            rm_xml = False if self.xml else True
            if failed_happen:  # failed files in non-interactive mode
                exit_message = "Some files failed downloading"
                rm_cookie = True
            else:
                exit_message = "Exit."
                rm_cookie = False

        exit_code = 1 if failed_happen else 0

        self._clean_exit(
            exit_message=exit_message,
            exit_code=exit_code,
            rm_cookie=rm_cookie,
            rm_xml=rm_xml,
        )

    def _byte_convert(self, byte_size):
        """
        Converts a number of bytes to a more human-readable
        format.

        """
        # Calculate and display total size of selected data
        adjusted = byte_size / (1024 * 1024)  # bytes to MB
        if adjusted < 1:
            adjusted = byte_size / 1024
            unit = "KB"
        elif adjusted < 1024:
            unit = "MB"
        else:
            adjusted /= 1024
            unit = "GB"
        size_string = "{:.2f} {}".format(adjusted, unit)
        return size_string

    def _clean_exit(self, exit_message=None, exit_code=0, rm_cookie=False, rm_xml=False):
        """
        Perform a sys.exit() while removing temporary files and
        informing the user.

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_clean_exit",
        )
        to_remove = []

        if rm_xml:
            xml_file = self.config.FILENAME_TEMPLATE_XML.format(self.query_info)
            to_remove.append(xml_file)
            logger.info(f"Removing xml @ {xml_file}")
        if rm_cookie:
            to_remove.append(self.config.FILENAME_COOKIE)
            logger.info(f"Removing cookie @ {self.config.FILENAME_COOKIE}")
        for f in to_remove:
            try:
                os.remove(f)
            except OSError:
                continue

        if exit_message:
            logger.info(exit_message)
        sys.exit(exit_code)

    def _check_for_xml(self, filename):
        """
        Uses hex code at the beginning of a file to try to determine if it's an
        XML file or not. This seems to be occasionally necessary; if pulling
        files from JGI tape archives, the server may error out and provide an
        XML error document instead of the intended file. This function should
        return False on all downloaded files, although false positives have not
        been thoroughly investigated.

        Adapted from http://stackoverflow.com/a/13044946/3076552

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_check_for_xml",
        )
        xml_hex = "\x3c"  # hex code at beginning of XML file
        read_length = len(xml_hex)
        with open(filename) as f:
            try:
                file_start = f.read(read_length)

                if file_start.startswith(xml_hex):  # XML file
                    logger.debug("cheker for xml: XML file")
                    return True
                else:  # hopefully all other file types
                    logger.debug("cheker for xml: hopefully all other file types")
                    return False
            except UnicodeDecodeError:  # compressed file
                logger.debug("cheker for xml: compressed file")
                return False

    def _check_md5(self, filename, md5_hash):
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_check_md5",
        )
        if not md5_hash:
            logger.warning(f"No MD5 hash listed for {filename}; skipping check")
            ret_val = True
        else:
            file_md5 = self._get_md5(filename)
            if file_md5 == md5_hash:
                logger.debug(f"MD5 hashes match for {filename} ({md5_hash})")
                ret_val = True
            else:
                logger.error(
                    f"MD5 hash mismatch for {filename} (local: {file_md5}, remote: {md5_hash})"
                )
                ret_val = False
        return ret_val

    def _check_sizeInBytes(self, filename, sizeInBytes):
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_check_sizeInBytes",
        )
        if not sizeInBytes:
            logger.info(f"No sizeInBytes listed for {filename}; skipping check")
            ret_val = True
        else:
            file_size_in_bytes = self._get_sizeInBytes(filename)
            if file_size_in_bytes == sizeInBytes:
                logger.debug(f"sizeInBytes match for {filename} ({sizeInBytes})")
                ret_val = True
            else:
                logger.error(
                    f"sizeInBytes mismatch for {filename} (local: {file_size_in_bytes}, remote: {sizeInBytes})"
                )
                ret_val = False
        return ret_val

    def _decision_tree(self):
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_decision_tree",
        )
        # Decision tree depending on if non-interactive options given
        regex_filter = None
        user_choice = None
        self.interactive = True

        if self.all_get:  # 参数指定get all
            # non-interactive
            user_choice = "a"  # "all_get mode"
            self.interactive = False
        elif self.regex:  # 参数指定regex
            # non-interactive
            user_choice = "r"  # "regex mode"
            regex_filter = self.regex
            self.interactive = False
            # non-interactive
        elif self.log_fails is not None:  # 参数指定
            user_choice = "l"  # "re-download log_fails mode"
            self.interactive = False

        logger.debug(f"regex_filter = {regex_filter}")
        logger.debug(f"user_choice = {user_choice}")
        logger.debug(f"interactive_and_display_info = {self.interactive}")
        logger.debug(
            f"QUERY RESULTS FOR '{str(self._desired_categories)[:400]}...(omit)...'\n"
        )

        for query_cat, v in sorted(
                iter(self._desired_categories.items()), key=lambda k_v: k_v[1]["catID"]
        ):
            print_list = []

            if not v["results"]:
                continue

            catID = v["catID"]

            self._dict_to_get[catID] = {}
            print_list.append(f" {catID}: {query_cat} ".center(80, "="))
            results = v["results"]

            for sub_cat, items in sorted(
                    iter(results.items()),
                    key=lambda sub_cat_items: (sub_cat_items[0], sub_cat_items[1]),
            ):
                print_list.append("{}:".format(sub_cat))
                for index, i in sorted(items.items()):
                    url = i["url"]
                    self._dict_to_get[catID][index] = url
                    if "md5" in i:
                        self._url_to_validate[url]["md5"] = i["md5"]
                    # the following elif takes care of MD5 > sizeInBytes rank-order
                    # in downstream processing
                    if "sizeInBytes" in i:
                        self._url_to_validate[url]["sizeInBytes"] = int(
                            i["sizeInBytes"]
                        )
                    print_index = " {}:[{}] ".format(str(catID), str(index))
                    date = self._format_timestamp(i["timestamp"])
                    date_string = "{:02d}/{}".format(date.tm_mon, date.tm_year)
                    size_date = "[{}|{}]".format(i["size"], date_string)
                    filename = i["filename"]
                    margin = 80 - (len(size_date) + len(print_index))
                    file_info = filename.ljust(margin, "-")
                    print_list.append("".join([print_index, file_info, size_date]))
            if self.interactive:
                print("\n".join(print_list))
                print()  # padding

        if not user_choice:
            # Ask user which files to download from xml
            user_choice = self._get_user_choice()

        if user_choice == "r":  # regex-based filename matching
            regex_filter = self._get_regex()
        # special case for downloading all available files
        # or filtering with a regular expression
        if user_choice in (  # non interactive
                "a",  # all_get mode
                "r",  # regex mode
                "l",  # re-download log_fails mode
        ):
            for k, v in sorted(self._dict_to_get.items()):
                for url in v.values():
                    # logger.debug(f'deal with v.values() = {str(v.values())[:200]}...(omit)...')
                    if regex_filter:
                        # fn = re.search(r".+/([^\/]+$)", u).group(1)
                        fn = re.search(r".+/([^/]+$)", url).group(1)
                        match = regex_filter.search(fn)
                        if not match:
                            # 匹配失败
                            self._urls_to_get = set(self._urls_to_get)
                            self._urls_to_get.discard(url)
                            continue  # 进入下一个文件进行判断
                    elif user_choice == "l" and url not in self._log_fails_loaded:
                        continue  # 无需添加此文件，进入下一个文件进行判断
                    # final add to download
                    self._urls_to_get = set(self._urls_to_get)
                    self._urls_to_get.add(url)  # self._urls_to_get is a set()
        else:
            # interactive
            # 选2:3-5这种
            # Retrieve user-selected file urls from dict
            self._parse_selection(user_choice)  # update self._selections
            for k, v in sorted(self._selections.items()):
                for i in v:
                    self._urls_to_get.add(self._dict_to_get[k][i])

        # logger.debug('update self._dict_to_get, self._url_to_validate')
        # logger.debug(f'self._urls_to_get = {str(self._urls_to_get)[:1000]}...(omit)...')
        # logger.debug(f'self._dict_to_get = {str(self._dict_to_get)[:1000]}...(omit)...')

    def _decompress_files(self, local_file_list, keep_original=False):
        """
        Decompresses list of files, and deletes compressed
        copies unless <keep_original> is True.

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_decompress_files",
        )
        logger.debug("decompress_files...")
        for f in local_file_list:
            self._extract_file(f, keep_original)

    def _download_from_url(self, url):
        """
        Attempts to download a file from JGI servers using cURL.

        Returns a tuple of (filename, cURL command used, success boolean)

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_download_from_url",
        )
        url_no_prefix = url.replace("&amp;", "&")  # for query local xml info

        # get md5 value for this file from xml
        # logger.debug(f'self._url_to_validate = {self._url_to_validate}')
        md5_hash = self._url_to_validate[url_no_prefix].get("md5", None)  # local xml
        # get sizeInBytes (local xml) and file_size (internet) for size checking
        sizeInBytes = self._url_to_validate[url_no_prefix].get(
            "sizeInBytes", None
        )  # local xml
        # counter
        error_counter = 0
        # download status
        loop = True
        success = False
        # filename to write in, parse from last of url
        filename = re.search(".+/(.+$)", url).group(1)

        while True:  # pass when loop == False
            if not loop:
                break

            if error_counter >= self.nretry:
                logger.critical(
                    f"error_counter is reaching to max size -> self.nretry = {self.nretry}"
                )
                # FUTURE, 支持了断点续传再改回来
                # logger.warning(
                #     'you can rerun your command and the unfinished file is needed for continuing download')
                success = False
                loop = False
                continue

            # temp_stemp_sizeize = 0  # 获取已写入的字节 temp_size
            if os.path.exists(filename):  # filename exists
                # check md5 for filename
                if not self._is_broken(filename, md5_hash=md5_hash, sizeInBytes=sizeInBytes):
                    logger.info(
                        f"File {filename} exists and passed md5 checking. Go on next task."
                    )
                    success = True
                    loop = False
                    continue
                else:  # failed
                    if os.path.exists(
                            filename + ".tmp"
                    ):  # filename.tmp exists at the sametime!
                        if not self._is_broken(
                                filename + ".tmp",
                                md5_hash=md5_hash,
                                sizeInBytes=sizeInBytes,
                        ):
                            logger.debug(
                                f"File {filename}.tmp passed md5 checking! Renaming"
                            )
                            os.remove(filename)
                            os.rename(filename + ".tmp", filename)
                            logger.debug(
                                f"File {filename} exists and passed md5 checking. Go on next task."
                            )
                            success = True
                            loop = False
                            continue
                        else:
                            # FUTURE: 由于JGI不支持断点续传，这个代码以后再用
                            # logger.error(f'File {filename} exists but is broken, file {filename}.tmp exists but is '
                            #              f'broken too. File {filename} is removed now.')
                            # os.remove(filename)
                            # temp_size = os.path.getsize(filename + '.tmp')
                            # logger.info(f'Try continuing download using {filename}.tmp. temp_size = {temp_size}')
                            # /FUTURE: 由于JGI不支持断点续传，这个代码以后再用
                            logger.warning(
                                f"File {filename} exists but is broken, file {filename}.tmp exists but is "
                                f"broken too."
                            )
                            logger.info(
                                "Because JGI does not support continuing download now, the two files will be removed"
                            )
                            os.remove(filename)
                            logger.debug(f"File {filename} is removed now.")
                            os.remove(filename + ".tmp")
                            logger.debug(f"File {filename}.tmp is removed now.")
                            success = False
                            loop = True
                            continue
                    else:
                        logger.warning(
                            f"File {filename} exists but is broken, file {filename}.tmp does not exists."
                            f"File {filename} is removed now."
                        )
                        os.remove(filename)
                        logger.debug("Try to start a new download task.")
                        success = False
                        loop = True
                        continue
            else:  # filename don't exist
                if os.path.exists(filename + ".tmp"):  # filename.tmp exists
                    # 如果.tmp文件已存在，则打开文件并获取已写入的字节 temp_size
                    # print(f'found tmp file for {filename}, attempt to continuing download')
                    # temp_size = os.path.getsize(filename + '.tmp')

                    logger.warning(
                        "Because JGI does not support continuing download now, the tmp file will be removed"
                    )
                    os.remove(filename + ".tmp")
                    logger.debug(f"File {filename}.tmp is removed now.")
            # set requests
            # https://dabing1022.github.io/2016/12/24/%E8%81%8A%E4%B8%80%E8%81%8AHTTP%E7%9A%84Range,%20Content-Range/
            if not url.startswith("http"):
                url = f"{self.config.URL_JGI_MAIN}{url}"
            logger.debug(f"download aim file from url = {url}")
            cookie_file = self.config.FILENAME_COOKIE
            headers = {"User-Agent": self._user_agent}
            with open(cookie_file, "rt") as f:
                cookies_dict = json.loads(f.read())
            cookies = requests.utils.cookiejar_from_dict(cookies_dict)

            # 想在Python中在不下载文件的情况下获取文件大小，可以使用requests库发送HEAD请求
            # HEAD请求不会下载文件，而是仅获取关于文件的一些元数据，如文件大小
            with requests.head(
                    url,
                    cookies=cookies,
                    stream=True,
                    headers=headers,
                    timeout=self.timeout,
                    proxies=self._proxies,
            ) as pre_response:
                if self._proxies:
                    logger.debug(f"requests.head using IP: {self._proxy_ip}")
                # check response status
                try:
                    pre_response.raise_for_status()  # 如果响应的状态码不是200，将引发HTTPError异常
                except HTTPError:
                    error_counter += 1
                    logger.warning(
                        "encounter HTTPError;"
                        f"error_counter = {error_counter}; pre_response.status_code = {pre_response.status_code}; "
                        "could not connect with server. Retry after 5s..."
                    )
                    time.sleep(5)
                    logger.info("start next loop")
                    loop = True
                    continue  # next try
                except Exception as e:
                    error_counter += 1
                    logger.warning(
                        f"encounter {e} (error);"
                        f"error_counter = {error_counter}; pre_response.status_code = {pre_response.status_code}; "
                        "unexpected error. Retry after 5s..."
                    )
                    time.sleep(5)
                    logger.info("start next loop")
                    success = False
                    loop = True
                    continue  # next try

                # get file size, if pre_response.header do not have it, replace it from sizeInBytes in xml file
                remote_file_size = int(pre_response.headers.get("Content-Length", 0))
                remote_file_size = (
                    remote_file_size if remote_file_size != 0 else sizeInBytes
                )
                # FUTURE: 由于JGI不支持断点续传，这个代码以后再用
                # if temp_size > remote_file_size:
                #     # error
                #     logger.error('local size = {temp_size} > remote_file_size = {remote_file_size}, nretry...')
                #     error_counter += 1
                #     os.remove(filename + '.tmp')
                #     success = False
                #     loop = True
                #     continue

            # core part for download
            # re-request use fixed headers
            # https://cizixs.com/2015/10/06/http-resume-download/
            # headers["Range"] = f"bytes={temp_size}-"
            # /FUTURE: 由于JGI不支持断点续传，这个代码以后再用
            with requests.get(
                    url,
                    cookies=cookies,
                    stream=True,  # 如果要下载大文件的话，就将steam设置为True，慢慢下载，而不是等整个文件下载完才返回。
                    headers=headers,
                    timeout=self.timeout,
                    proxies=self._proxies,
            ) as file_response:
                if self._proxies:
                    logger.debug(f"requests.get using IP: {self._proxy_ip}")
                # 分段下载
                # FUTURE: 由于JGI不支持断点续传，代码以后再测试
                # print(file_response.headers)
                # print(file_response.status_code)
                # ################### core ########################
                with open(filename + ".tmp", "ab") as f, tqdm(
                        desc=filename,
                        total=remote_file_size,
                        # initial=temp_size,  # FUTURE: 由于JGI不支持断点续传，这个代码以后再用
                        unit="iB",
                        unit_scale=True,
                        unit_divisor=1024,
                ) as pbar:
                    try:
                        for data in file_response.iter_content(chunk_size=1024):
                            data_size = f.write(data)
                            pbar.update(data_size)
                    except (
                            InvalidChunkLength or ProtocolError or ChunkedEncodingError
                    ) as e:
                        logger.error(f"Invalid chunk encoding {e}")
                        success = False
                        loop = True
                        continue
                # ################### /core ########################
                # finish download
                # start check
                logger.debug(f"File {filename}.tmp seems done, checking md5 value")
                # 如果filename md5通过则删除tmp
                if not self._is_broken(
                        filename + ".tmp",
                        md5_hash=md5_hash,
                        sizeInBytes=remote_file_size,
                ):
                    logger.debug(f"File {filename}.tmp passed md5 checking! Renaming")
                    os.rename(filename + ".tmp", filename)
                    logger.debug(
                        f"File {filename} passed md5 checking. Go on next task."
                    )
                    success = True
                    loop = False
                    continue  # next task
                else:
                    error_counter += 1
                    logger.info(
                        f"File {filename}.tmp is broken! Removing and retry after 2 seconds..."
                    )
                    time.sleep(2)
                    os.remove(filename + ".tmp")
                    success = False
                    loop = True
                    continue  # next try
        return filename, success

    def _download_list(self, url_list):
        """
        Attempts download command on a list of partial file
        URLs (completed by download_from_url()).

        Returns a list of successfully-downloaded files and a
        list of unsuccessful URLs

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_download_list",
        )
        # Run curl commands to retrieve selected files
        # Make sure the URL formats conforms to the Genome Portal format

        broken_urls = []
        broken_files = []
        start_time = time.time()
        for url in url_list:
            current_time = time.time()
            # refresh the session cookie every 5 minutes
            if current_time - start_time > 300:
                logger.info("refresh the session cookie every 5 minutes")
                logger.info("run self.login.")
                try:
                    os.remove(self.config.FILENAME_COOKIE)
                except Exception:
                    pass
                self._load_cookie()
                logger.info("login succeed.")
                start_time = time.time()

            fn, success = self._download_from_url(url)

            if not success:
                logger.warning(
                    f"File {fn} failed to download, appending to the file "
                    f"@ {self.config.FILENAME_TEMPLATE_LOG_FAIL.format(self.query_info)}. "
                    f"You can use parameter --log_fails to re-download these failed files "
                    f"after this run finish."
                )
                logger.warning(
                    "tips: it is possible that the md5 value on JGI is wrong. If retry failed too, you can "
                    "download it from the website manually!"
                )
                broken_urls.append(url)
                broken_files.append(fn)
            else:
                self._downloaded_files.append(fn)
        broken_urls = [
            u
            for u, f in zip(broken_urls, broken_files)
            if f not in self._downloaded_files
        ]
        logger.debug(
            f" self._download_list return broken_urls = {broken_urls}, broken_files = {broken_files}"
        )
        return broken_urls, broken_files

    def _extract_file(self, file_path, keep_compressed=False):
        """
        Native Python file decompression for tar.gz and .gz files.

        TODO: implement .zip decompression

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_extract_file",
        )
        tar_pattern = "tar.gz$"  # matches tar.gz
        gz_pattern = "(?<!tar)\.gz$"  # excludes tar.gz
        endings_map = {"tar": (tarfile, "r:gz", ".tar.gz"), "gz": (gzip, "rb", ".gz")}
        relative_name = os.path.basename(file_path)
        if re.search(tar_pattern, file_path):
            logger.info(f"tar.gz file decompression for {file_path}")
            opener, mode, ext = endings_map["tar"]
            with opener.open(file_path) as f:
                file_count = len(f.getmembers())
                if file_count > 1:  # make sub-directory to unpack into
                    dir_name = relative_name.rstrip(ext)
                    try:
                        os.mkdir(dir_name)
                    except FileExistsError:
                        pass
                    destination = dir_name
                else:  # single file, extract into working directory
                    destination = "."

                def is_within_directory(directory, target):
                    abs_directory = os.path.abspath(directory)
                    abs_target = os.path.abspath(target)

                    prefix = os.path.commonprefix([abs_directory, abs_target])

                    return prefix == abs_directory

                def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                    for member in tar.getmembers():
                        member_path = os.path.join(path, member.name)
                        if not is_within_directory(path, member_path):
                            raise Exception("Attempted Path Traversal in Tar File")

                    tar.extractall(path, members, numeric_owner=numeric_owner)

                safe_extract(f, destination)
        elif re.search(gz_pattern, file_path):
            logger.info(f"gz file decompression for {file_path}")
            opener, mode, ext = endings_map["gz"]
            # out_name = file_path.rstrip(ext)
            out_name = relative_name.rstrip(ext)
            with opener.open(file_path) as f, open(out_name, "wb") as out:
                for l in f:
                    out.write(l)
        else:
            logger.info(f"Skipped decompression for '{file_path}'")
            return
        if not keep_compressed:
            logger.info("remove compressed files")
            os.remove(file_path)

    def _format_found(self, d, filter_found=False):
        """
        Reformats the output from xml_hunt()

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_format_found",
        )
        logger.debug("get categories from config (including possible user additions)")
        output = {}
        for p, c in sorted(d.items()):
            layers = [e for e in p.split(":") if e]
            if filter_found:
                if not any(cat in layers for cat in self.config.info["categories"]):
                    continue
            if len(layers) == 1:
                top = parent = layers[0]
            else:
                top = layers[-2]  # either -2 or -1 works well, != parent
                parent = layers[-1]  # either -2 or -1 works well, != top
            if top not in output:
                output[top] = defaultdict(dict)
            if parent not in output[top]:
                output[top][parent] = c
            else:
                output[top][parent].extend(c)
        return output

    def _format_timestamp(self, time_string):
        """
        Parses the timestamp string from an XML document
        of the form "Thu Feb 27 16:38:54 PST 2014"
        and returns a string of the form "2014".

        """
        # Remove platform-dependent timezone substring
        # of the general form "xxT"
        tz_pattern = re.compile("\s[A-Z]{3}\s")
        time_string = tz_pattern.sub(" ", time_string)
        # get the desired time info
        time_info = time.strptime(time_string, "%a %b %d %H:%M:%S %Y")
        # year = str(time_info.tm_year)
        return time_info

    def _get_sizes(self, d, sizes_by_url=None):
        """
        Builds a dictionary of url:sizes from
        output of get_file_list()

        """
        for k, v in d.items():
            if isinstance(v, dict):
                if "url" in v:
                    address = v["url"]
                    try:
                        size = int(v["sizeInBytes"])
                    except:
                        size = None
                    sizes_by_url[address] = size
                else:
                    self._get_sizes(v, sizes_by_url)
        return sizes_by_url

    def _get_md5(self, *fns, buffer_size=65536):
        hash = md5()
        for fn in fns:
            with open(fn, "rb") as f:
                while True:
                    data = f.read(buffer_size)
                    if not data:
                        break
                    hash.update(data)

        return hash.hexdigest()

    def _get_sizeInBytes(self, filename):
        try:
            file_size_in_bytes = os.path.getsize(filename)
        except:
            file_size_in_bytes = 0

        return file_size_in_bytes

    def _get_regex(self):
        """
        Get regex pattern from user, compile and return.

        """
        # TODO make this exit gracefully if user can't
        # manage to get a working regex
        compile_success = False
        pattern = ""
        while compile_success is False:
            if self.interactive and not self.regex:
                pattern = input("Regex pattern: ")
            else:
                pattern = self.regex
            try:
                pattern = re.compile(pattern)
                compile_success = True
                print(f"Your pattern = {pattern}")
            except:
                print("[!] ERROR: Regex pattern failed to compile.")

        return re.compile(pattern)

    def _get_user_choice(self):
        """
        Get user file selection choice(s)

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_get_user_choice",
        )
        choice = input(
            "Enter file selection ('q' to quit, "
            "'u' to review syntax/usage, 'a' for all, "
            "'r' for regex-based filename matching):\n> "
        )
        logger.debug(f"choice = {choice}")
        if choice == "u":
            print()
            print(JGIDoc.select_blurb)
            print()
            return self._get_user_choice()
        elif choice == "a":
            return choice
        elif choice == "r":
            return choice
        elif choice == "q":
            keep_temp = input(
                f"Keep temporary files ('{self.config.FILENAME_TEMPLATE_XML.format(self.query_info)}' "
                f"and '{self.config.FILENAME_COOKIE}')? (y/n): "
            )
            if keep_temp.lower() in ("y", "yes"):
                rm_cookie = False
                rm_xml = False
            else:
                rm_cookie = True
                rm_xml = True
            self._clean_exit(
                exit_message="User choose to exit",
                exit_code=0,
                rm_cookie=rm_cookie,
                rm_xml=rm_xml,
            )
        else:
            return choice

    def _is_broken(self, filename, min_size_bytes=20, md5_hash=None, sizeInBytes=None):
        """
        Rudimentary check to see if a file appears to be broken.

        filename without ".tmp"
        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_is_broken",
        )
        if (
                not os.path.isfile(filename)
                or os.path.getsize(filename) < min_size_bytes
                or (
                self._check_for_xml(filename)
                and not filename.lower().endswith("xml")
        )
                or (
                not self._check_md5(filename, md5_hash)
                or not self._check_sizeInBytes(filename, sizeInBytes)
        )
        ):
            logger.debug("File is broken.")
            return True
        else:
            logger.debug("File is intact.")
            return False

    def _parse_selection(self, user_input):
        """
        Parses the user choice string and returns a dictionary
        of categories (keys) and choices within each category
        (values).

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_parse_selection",
        )
        parts = user_input.split(";")
        for p in parts:
            if len(p.split(":")) > 2 or p.count(":") == 0:
                logger.error(f"Can't parse desired input\n?-->'{p}'")
                user_choice = self._get_user_choice()
                self._parse_selection(user_choice)
            category, indices = p.split(":")
            category = int(category)
            self._selections[category] = []
            indices = indices.split(",")
            for i in indices:
                try:
                    self._selections[category].append(
                        int(i)
                    )  # if it's already an integer
                except ValueError:
                    try:
                        start, stop = list(map(int, i.split("-")))
                    except:
                        logger.critical(f"can't parse desired input\n?-->'{i}'")
                        self._clean_exit(
                            exit_message="exit with error",
                            exit_code=1,
                            rm_cookie=False,
                            rm_xml=False if self.xml else True,
                        )

                    add_range = list(range(start, stop + 1))

                    for e in add_range:
                        self._selections[category].append(e)

    def _xml_hunt(self, xml_file):
        """
        Gets list of all XML entries with "filename" attribute,
        and returns a dictionary of the file attributes keyed
        by a ":"-joined string of parent names.

        """
        logger = get_logger(
            level=self.log_level,
            module_name=__module_name__,
            func_name="_xml_hunt",
        )
        logger.debug("xml_hunt...")
        root = ET.iterparse(xml_file, events=("start", "end"))
        parents = []
        matches = {}
        try:
            for event, element in root:
                if element.tag not in ["folder", "file"]:  # skip topmost categories
                    continue
                if element.tag == "folder":
                    if event == "start":  # add to parents
                        parents.append(element.attrib["name"])
                    elif event == "end":  # strip from parents
                        del parents[-1]
                    continue
                if event == "start" and element.tag == "file":
                    parent_string = ":".join(parents)
                    try:
                        matches[parent_string].append(element.attrib)
                    except KeyError:
                        matches[parent_string] = [element.attrib]
        except ET.ParseError as e:
            logger.critical(
                "Query or parse xml failed, start to remove cookie. Please try again. "
                "If it is still failed, maybe the username / password is wrong? "
                "Try to use --overwrite_conf parameter to re-login"
            )
            try:
                os.remove(self.config.FILENAME_COOKIE)
            except Exception:
                pass
            sys.exit(f"Exit with error {e}")
        return matches

    def _uniqueify(self, children):
        """
        Takes a list of child XML elements (dicts of attribs) as
        returns a filtered list of only unique filenames for a given
        month/year timestamp (e.g. duplicates are allowed if month/year
        is different).

        """
        unique = {}
        for child in children:
            try:
                fn = child["filename"]
                date = self._format_timestamp(child["timestamp"])
                date_string = (date.tm_mon, date.tm_year)
                uid = (fn, date_string)
            except KeyError:
                continue
            if fn not in unique:
                unique[uid] = child
            else:
                existing = unique[uid].get("fileType", None)
                if existing == "Unknown":
                    existing = None
                current = child.get("fileType", None)
                if current == "Unknown":
                    current = None
                if current is not None and existing is None:
                    unique[uid] = child

        return unique.values()



"""Module for logging

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

Logging format and color

example 1: # TODO
    bioat list
        <in shell>:
            $ bioat list
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> bioat.list()
            >>> print(bioat.list())

example 2:
    _example_
"""

import logging
import logging.handlers
import sys
from logging import CRITICAL, DEBUG, ERROR, INFO, NOTSET, WARNING, Logger

import coloredlogs

# !dont move this -> __PKG_NAME__, referenced by many functions
__PKG_NAME__ = "bioat"

LEVEL = dict(
    zip(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        [CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET],
    )
)


def get_logger(
    level="ERROR",
    module_name=__PKG_NAME__,
    func_name: None | str = None,
) -> Logger:
    """Return a logger object.

    :param level: logger level, can be set to CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET, defaults to "ERROR"
    :type level: str, optional
    :param module_name: _description_, defaults to __PKG_NAME__
    :type module_name: _type_, optional
    :param func_name: _description_, defaults to None
    :type func_name: _type_, optional
    :return: a logger object
    :rtype: Logger
    """
    level = LEVEL[level.upper()]

    # define the name var in fmt: %(name)
    if func_name:
        logger = logging.getLogger(f"{module_name}.{func_name}")
    else:
        logger = logging.getLogger(f"{module_name}")
    # 设置日志格式
    fmt = "%(asctime)s - [%(name)s] - %(filename)s[line:%(lineno)4d] - %(levelname)+8s: %(message)s"
    # add color feature to logger obj (cli color)
    coloredlogs.DEFAULT_FIELD_STYLES = dict(
        asctime=dict(color="green"),
        name=dict(color="blue"),
        filename=dict(color="magenta"),
        lineno=dict(color="cyan"),
        levelname=dict(color="blue", bold=True),
    )
    coloredlogs.install(fmt=fmt, level=level, logger=logger, reconfigure=True)
    return logger


if __name__ == "__main__":
    logger = get_logger(
        "DEBUG",
        # module_name="bioat.logger",
        # func_name="test_function"
    )
    logger.debug("Python debug")
    logger.info("Python info")
    logger.warning("Python warning")
    logger.error("Python Error")
    logger.critical("Python critical")



"""Doc.
dataclasses:
    - https://docs.python.org/zh-cn/3/library/dataclasses.html
    - python version >= 3.7.0, PEP 557, PEP 526
"""
from dataclasses import dataclass


@dataclass
class Bed:
    chromosome: str
    start: int
    end: int
    name: str
    score: int
    strand: str


@dataclass
class Bam:
    pass


@dataclass(frozen=True)
class Fasta:
    """
    usage:
        fa = Fasta('this_header', 'AGCTACGTCCCCTGA')
        fa
        print(fa)
    """
    header: str
    sequence: str

    def __eq__(self, other):
        if self.header == other.header and self.sequence == other.sequence:
            return True
        else:
            return False

    def __len__(self):
        return len(self.sequence)

    @property
    def length(self):
        return len(self.sequence)

    def __str__(self, line_width: int = 20):
        ls = []

        for i in range(self.length // line_width + 1):
            ls.append(self.sequence[i * line_width:(i + 1) * line_width])
        s = '\n'.join(ls)
        s = f'>{self.header}\n{s}'
        return s

    def __repr__(self, show_all: bool = False, show_length: int = 10):
        if show_all:
            return f"Fasta(header='{self.header}', sequence='{self.sequence}')"
        else:
            if self.length > show_length:
                s = self.sequence[: show_length // 2] + '...' + self.sequence[-show_length // 2:]
                return f"Fasta(header='{self.header}', sequence='{s}')"
            else:
                return f"Fasta(header='{self.header}', sequence='{self.sequence}')"


@dataclass(frozen=True)
class Fastq(Fasta):
    header: str
    sequence: str
    extras: str
    quality: str

    def __eq__(self, other, check_quality: bool = False, check_extras: bool = False):
        # rewrite from class Fasta
        checker = [self.sequence == other.sequence]

        if check_extras:
            checker.append(self.extras == other.extras)
        if check_quality:
            checker.append(self.quality == other.quality)

        if all(checker):
            return True
        else:
            return False

    def __str__(self, line_width: int = 20):
        # rewrite from class Fasta
        return f'@{self.header}\n{self.sequence}\n{self.extras}\n{self.quality}\n'

    def __repr__(self):
        # rewrite from class Fasta
        return (f"Fastq(header='{self.header}', sequence='{self.sequence}', "
                f"extras='{self.extras}', quality='{self.quality}')")

    def is_valid(self):
        if self.length == len(self.quality):
            return True
        else:
            return False

    # def get_phread_info(self):
    #     # 判断是Phred33还是Phred64
    #     # https://www.jianshu.com/p/248308513e2e
    #     pass


@dataclass(frozen=True)
class Assembly(object):
    contigs: list[Fasta]
    length: int
    path: str

    def __init__(self, path: str):
        pass


@dataclass
class VCF:
    pass



"""_summary_

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

_description_

example 1:
    bioat list
        <in shell>:
            $ bioat list
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> bioat.list()
            >>> print(bioat.list())

example 2:
    _example_
"""

import sys

import pandas as pd

from bioat.logger import get_logger

__all__ = ['set_option']
__module_name__ = 'bioat.lib.libpandas'


def set_option(
    progres_bar: bool = True,
    max_colwidth: int = 40,
    display_width: int = 120,
    display_max_columns: int | None = None,
    display_max_rows: int = 50,
    log_level="INFO",
):
    logger = get_logger(
        level=log_level, module_name=__module_name__, func_name="set_option"
    )
    # 通过判断是否在notebook环境中，来选择不同的进度条
    try:
        from IPython.display import display

        logger.info("set pandas from a notebook environment")
        from tqdm.notebook import tqdm
    except ImportError:
        logger.info("set pandas from a terminal environment")
        from tqdm import tqdm

    if progres_bar:
        logger.info("set pandas: tqdm.pandas")
        tqdm.pandas()
    logger.info(f'set pandas: max_colwidth={max_colwidth}')
    pd.set_option("max_colwidth", max_colwidth)  # column最大宽度
    logger.info(f'set pandas: display.width={display_width}')
    pd.set_option("display.width", display_width)  # dataframe宽度
    logger.info(f'set pandas: display.max_columns={display_max_columns}')
    pd.set_option("display.max_columns", display_max_columns)  # column最大显示数
    logger.info(f'set pandas: display.max_rows={display_max_rows}')
    pd.set_option("display.max_rows", display_max_rows)  # row最大显示数
    return


def move_column(df: pd.DataFrame, to_move: str, after: str | None = None, before: str | None = None,
                first: bool | None = None, last: bool | None = None, inplace=False) -> pd.DataFrame | None:
    """
    Move a specified column in a DataFrame to a new position.

    :param df: The DataFrame to operate on.
    :param to_move: The name of the column to move.
    :param after: (Optional) If provided, move the `to_move` column after the `after` column.
    :param before: (Optional) If provided, move the `to_move` column before the `before` column.
    :param first: (Optional) If True, move the column to the first position.
    :param last: (Optional) If True, move the column to the last position.
    :param inplace: (Optional) If True, modify the original DataFrame without returning a new one.
    :return: A new DataFrame with the moved column, or None if inplace=True.
    """
    # Validate inputs for to_move
    if to_move not in df.columns:
        raise ValueError(f"Column '{to_move}' not found in the DataFrame.")
    # Validate inputs for first and last
    if first and last:
        raise ValueError("Cannot specify both 'first' and 'last' as True. Please choose one.")
    if first is not None and not isinstance(first, bool):
        raise ValueError("'first' must be a boolean or None.")
    if last is not None and not isinstance(last, bool):
        raise ValueError("'last' must be a boolean or None.")
    # Validate inputs for before and after
    if after is not None:
        if after not in df.columns:
            raise ValueError(f"Column '{after}' not found in the DataFrame. Cannot use it as a reference.")
        if to_move == after:
            raise ValueError(f"to_move({to_move}) is not allowed to be the same with after({after})")
    if before is not None:
        if before not in df.columns:
            raise ValueError(f"Column '{before}' not found in the DataFrame. Cannot use it as a reference.")
        if to_move == before:
            raise ValueError(f"to_move({to_move}) is not allowed to be the same with before({before})")
    if after is not None and before is not None:
        raise ValueError("Cannot specify both 'after' and 'before'. Please choose one.")

    current_columns = df.columns.tolist()
    if first:
        new_columns = [to_move] + [col for col in current_columns if col != to_move]
    elif last:
        new_columns = [col for col in current_columns if col != to_move] + [to_move]
    else:
        # Find the index of the column to move
        index_to_move = current_columns.index(to_move)
        # Remove the column from its current position
        current_columns.pop(index_to_move)
        # Determine the new position
        if after is not None:
            index_after = current_columns.index(after) + 1
            new_columns = current_columns[:index_after] + [to_move] + current_columns[index_after:]
        elif before is not None:
            index_before = current_columns.index(before)
            new_columns = current_columns[:index_before] + [to_move] + current_columns[index_before:]
        else:
            raise ValueError("you must define one of the params(first/last/before/after) for operation")

    # Return the modified DataFrame or None if inplace=True
    if inplace:
        df[current_columns] = df[new_columns]
        return None
    else:
        df = df[new_columns].copy()
        return df


if __name__ == '__main__':
    set_option()

    data = pd.DataFrame(
        {
            'a': [1, 2, 3, 4],
            'b': [2, 2, 3, 4],
            'c': [3, 2, 3, 4],
            'd': [4, 2, 3, 4],
        }
    )
    # print(move_column(data, to_move='d'))  # ValueError: you must define one of the params(first/last/before/after)
    # for operation
    # print(move_column(data, to_move='a', before='a'))  # to_move(a) is not allowed to be the same with before(a)
    # print(move_column(data, to_move='a', before='b'))
    # print(move_column(data, to_move='a', before='c'))
    # print(move_column(data, to_move='a', before='d'))
    # print(move_column(data, to_move='d', after='a'))
    # print(move_column(data, to_move='d', after='b'))
    # print(move_column(data, to_move='d', after='c'))
    # print(move_column(data, to_move='d', after='d'))  # to_move(d) is not allowed to be the same with after(d)
    # print(move_column(data, to_move='d', before='a'))
    # print(move_column(data, to_move='d', before='b'))
    # print(move_column(data, to_move='d', before='c'))
    # print(move_column(data, to_move='d', before='d'))  # to_move(d) is not allowed to be the same with before(d)
    # print(move_column(data, 'b', first=True))
    # print(move_column(data, to_move='b', last=True))
    print(f'data = \n{data}, \nid = {id(data)}')
    # data_new = move_column(data, to_move='a', last=True, inplace=False)
    data_new = move_column(data, to_move='a', last=True, inplace=True)
    print(f'data_new = \n{data_new}, \nid = {id(data_new)}')
    print(f'data = \n{data}, \nid = {id(data)}')



from bioat.lib.libpath import HOME, exists_in_PATH
from bioat.lib.libfastx import cas_finder, cas13_finder


class CrisprTools:
    """CRISPR mining toolbox."""

    # for development test
    _prodigal = (
        "prodigal"
        if exists_in_PATH("prodigal")
        else f"{HOME}/micromamba/envs/snakepipes_Cas-mining/bin/prodigal"
    )
    _pilercr = (
        "pilercr"
        if exists_in_PATH("pilercr")
        else f"{HOME}/micromamba/envs/snakepipes_Cas-mining/bin/pilercr"
    )

    # /for development test

    def cas_finder(
            self,
            input_fa,
            output_faa=None,
            output_contig_fa=None,
            output_crispr_info_tab=None,
            lmin=3000,  # 3001 in Nature Methods paper
            lmax=None,
            extend=10_000,
            temp_dir=None,
            prodigal=_prodigal,
            prodigal_mode='meta',
            pilercr=_pilercr,
            rm_temp=True,
            log_level="INFO"
    ):
        """De novo annotation for Cas candidates from neighbor of CRISPR loci.

        :param input_fa: metagenome.fa (with many contigs in it)
        :param output_faa: De novo annotated Cas candidates
        :param output_contig_fa: De novo annotated Cas candidate whole contigs
        :param output_crispr_info_tab: De novo annotated CRISPR info table (csv)
        :param lmin: min length for a contig
        :param lmax: max length for a contig
        :param extend: proteins are considered over how far from the start/end of the CRISPR loci
        :param temp_dir: folder to put temp files in
        :param prodigal: the executable prodigal path
        :param pilercr: the executable pilercr path
        :param prodigal_mode: meta | single, mode for prodigal annotation
        :param rm_temp: set False to keep the temp files
        :param log_level: set DEBUG to see output from prodigal and pilercr
        """
        cas_finder(
            input_fa=input_fa,
            output_faa=output_faa,
            output_contig_fa=output_contig_fa,
            output_crispr_info_tab=output_crispr_info_tab,
            lmin=lmin,
            lmax=lmax,
            extend=extend,
            temp_dir=temp_dir,
            prodigal=prodigal,
            prodigal_mode=prodigal_mode,
            pilercr=pilercr,
            rm_temp=rm_temp,
            log_level=log_level,
        )

    # def cas12_finder(self):
    #     """Cas12 mining toolbox.
    #
    #     Under development!
    #     """
    #     pass

    def cas13_finder(
            self, input_faa, output_faa=None, lmin=200, lmax=1500, log_level="INFO"
    ):
        """De novo annotation for Cas13 candidates from proteins.faa.

        :param input_faa: cas_candidates.faa
        :param output_faa: cas13_candidates.faa
        :param lmin: min length for a cas candidate
        :param lmax: max length for a cas candidate
        :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
        """
        """Cas13 mining toolbox."""
        # start to test HEPN filter
        cas13_finder(
            input_faa=input_faa,
            output_faa=output_faa,
            lmin=lmin,
            lmax=lmax,
            log_level=log_level,
        )


if __name__ == "__main__":
    pass



import sys

from bioat.lib.libjgi import JGIOperator
from bioat.logger import get_logger

__module_name__ = "bioat.metatools"


class MetaTools:
    """Metagenome toolbox."""

    def JGI_query(
        self,
        # pick one from three
        query_info: str | None = None,
        xml: str | None = None,
        log_fails: str | None = None,
        # runtime params
        nretry: int = 4,
        timeout: int = 60,
        regex: str | None = None,
        all_get: bool = False,
        overwrite_conf: bool = False,
        filter_files: bool = False,
        proxy_pool: str | None = None,
        just_query_xml: bool = False,
        # doc helper
        syntax_help: bool = False,
        usage: bool = False,
        # log
        log_level: str = "INFO",
    ):
        """JGI_query, a tool for downloading files from JGI-IMG database.

        This script will list and retrieve files from JGI using the curl API.
        It will return a list of all files available for download for a given query organism.

        The source code is adapted from https://github.com/glarue/jgi-query

        :param query_info: (input) organism name formatted per JGI's abbreviation.
            For example, 'Nematostella vectensis' is abbreviated by JGI as 'Nemve1'.
            The appropriate abbreviation may be found by searching for the organism on JGI;
            the name used in the URL of the 'Info' page for that organism is the correct abbreviation.
            The full URL may also be used for this argument.
        :param xml: (input) specify a local xml file for the query instead of retrieving a new copy from JGI
        :param log_fails: (input) nretry downloading from URLs listed in log file
        :param nretry: number of times to nretry downloading files with errors (0 to skip such files)
        :param timeout: timeout (seconds) for downloading, set -1 to disable this
        :param regex: (no interactive) Regex pattern to use to auto-select and download files
        :param all_get: (no interactive) Auto-select and download all files for query
        :param overwrite_conf: (interactive) initiate configuration dialog to overwrite existing user/password
            configuration
        :param filter_files: (work in progress) filter organism results by config categories instead of reporting all
            files listed by JGI for the query
        :param proxy_pool: str | None = None, http://abc.com:port, see https://github.com/hermanzhaozzzz/proxy_pool
        :param just_query_xml: bool, set True if you just want save XML file.
        :param syntax_help: (doc mode) syntax_help
        :param usage: (doc mode) print verbose usage information and exit
        :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
        """
        # load or create JGI account info &
        # auto check if you need overwrite user info
        operator = JGIOperator(
            query_info=query_info,
            xml=xml,
            log_fails=log_fails,
            nretry=nretry,
            timeout=timeout,
            regex=regex,
            all_get=all_get,
            overwrite_conf=overwrite_conf,
            filter_files=filter_files,
            proxy_pool=proxy_pool,
            just_query_xml=just_query_xml,
            syntax_help=syntax_help,
            usage=usage,
            log_level=log_level,
        )
        logger = get_logger(
            level=log_level,
            module_name=__module_name__,
            func_name="JGI_query",
        )
        logger.debug("run query")
        operator.query()
        logger.debug("parse xml to json")
        operator.parse_xml()
        logger.debug(
            "start to download; calculate and display total size of selected data"
        )
        operator.download()

    def mafft(self):
        pass



import gzip
import sys

import pandas as pd

from bioat.exceptions import BioatFileFormatError, BioatFileNotCompleteError
from bioat.lib.libfastx import format_this_fastx
from bioat.logger import get_logger

__module_name__ = "bioat.fastxtools"


class FastxTools:
    """FASTA & FASTQ toolbox."""

    def __init__(self):
        self.fastx = None
        pass

    def fmt_this(self, file: str, new_file: str | None = None, log_level="WARNING"):
        """format fasta file and make it easier to see.

        :param file: input filename, fasta
        :param new_file: output filename, default is None, and when it is None, `file` will be replaced
        :param log_level: like others
        """
        format_this_fastx(old_file=file, new_file=new_file, log_level=log_level)

    def mgi_parse_md5(self, file: str, log_level="WARNING"):
        """Read mgi-like md5 file and convert to a normal md5 file.

        :param file: file name of a mgi-like md5 file.
        :param log_level: INFO/DEBUG/WARNING/ERROR, default is WARNING.
        """
        logger = get_logger(
            level=log_level,
            module_name=__module_name__,
            func_name="mgi_parse_md5",
        )
        try:
            df = pd.read_csv(
                file,
                header=None,
                index_col=False,
                sep="  ",
                engine="python",
            )
            df.columns = ["md5", "filename"]
            df.md5 = df.md5.map(lambda x: x.lower())
        except ValueError as e:
            logger.error(BioatFileFormatError(e))
            exit(1)
        df.filename = df.filename + ".gz"
        to_path = file.replace(".txt", "") + ".fix.md5"
        logger.info(f"write to {to_path}")
        df.to_csv(to_path, header=False, index=False, sep="\t")

    def check_completeness(self, file: str, fmt="FASTQ", log_level="WARNING"):
        """正在开发中

        :param file: :param str file: path of input <fastq | fastq.gz | fastx | fastx.gz>
        :param fmt: FASTQ | FASTA
        :param log_level:
        :return: str, PASS | FILENAME_LOG_FAIL
        """
        logger = get_logger(
            level=log_level,
            module_name=__module_name__,
            func_name="check_completeness",
        )
        pass

    @staticmethod
    def _load_fastx_generator(file, log_level="WARNING"):
        """

        :param str file: path of input <fastq | fastq.gz | fastx | fastx.gz>
        :return: a generator for all reads:
            i.e. print(next(obj)) -> ['header', 'seq', 'info', 'quality']
        :rtype: generator
        """
        logger = get_logger(
            level=log_level,
            module_name=__module_name__,
            func_name="_load_fastx_generator",
        )
        f = open(file, "rt") if not file.endswith(".gz") else gzip.open(file, "rt")
        # FASTQ @  | FASTA >
        symbol = f.read(1)
        f.close()
        f = open(file, "rt") if not file.endswith(".gz") else gzip.open(file, "rt")

        if symbol == "@":
            logger.debug("detect FASTQ file")
            read = []
            line = f.readline().rstrip()

            while True:
                if not line:
                    break
                else:
                    if line.startswith("@"):
                        # header or not complete
                        n = len(read)
                        if n == 0:
                            # header, 第一次循环开始
                            read.append(line)  # add header!
                            line = f.readline().rstrip()  # next line
                            continue
                        elif n == 4:
                            # read == ['header', 'seq', 'info', 'quality']
                            yield read
                            read = []
                            read.append(line)
                            line = f.readline().rstrip()
                        elif n == 3:
                            """
                            @@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6 # line! 期望它是 header！现在它是 quality
                            [
                                '@Beta12AdemL1C001R00100001768/1',
                                'ATCCCCGTATCTTCACCCCACCACAAACTATTAG',
                                '+',
                            ]'@@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6'

                            """
                            read.append(line)
                            line = f.readline().rstrip()

                            if not line.startswith("@"):
                                raise ValueError("The file may be incomplete!")
                        else:
                            f.close()
                            raise ValueError("The file may be incomplete!")

                        # header line!
                        # read.append(line)  # add header !
                    else:
                        # not header line!
                        read.append(line)
                        line = f.readline().rstrip()

            # ls.append(read)
            yield read
            f.close()
            # return ls

        elif symbol == ">":
            logger.debug("detect FASTA file")
            read = []
            seq = ""
            line = f.readline().rstrip()

            while True:
                if not line:
                    f.close()
                    break
                else:
                    if line.startswith(">"):
                        # 读取 header！
                        n = len(read)

                        if n == 0:
                            # 第一次循环
                            read.append(line)
                            line = f.readline().rstrip()
                        elif n == 1:
                            # 已经有一个 header 了！现在缺 seq
                            read.append(seq)  # add seq line
                            # ls.append(read)
                            yield read
                            read = []  # 重置 read 这个 list
                            read.append(line)
                            line = f.readline().rstrip()
                            seq = ""  # 重置 seq 这个 str
                        else:
                            f.close()
                            logger.error(
                                BioatFileNotCompleteError("The file may be incomplete!")
                            )
                            exit(1)
                    else:
                        # 读取并添加 seq
                        seq += line
                        line = f.readline().rstrip()

            read.append(seq)
            # ls.append(read)
            yield read
            # return ls
        else:
            f.close()
            logger.error(
                BioatFileFormatError(
                    "Input line one must starts with `@` for FASTQ or `>` for FASTA!"
                )
            )
            exit(1)
        f.close()

    # def get_headers(self):
    #     # if self.fastx is None:
    #     self.fastx = self._load_fastx_generator()

    #     return [i[0][1:] for i in self.fastx]

    # def get_sequence(self):
    #     # if self.fastx is None:
    #     self.fastx = self._load_fastx_generator()

    #     return [i[1] for i in self.fastx]

    # def get_record_dict(self):
    #     # if self.fastx is None:
    #     self.fastx = self._load_fastx_generator()

    #     return {i[0][1:]: i[1] for i in self.fastx}

    # cli subcmd
    def filter_read_contains_N(self, file: str, output=sys.stdout.name):
        """Filter read contains N base in FASTA or FASTQ



        :param file: FASTA | FASTQ file name.
        :param output: FASTA | FASTQ file name file, stdout if not assigned, format as input.
        """
        if self.fastx is None:
            self.fastx = self._load_fastx_generator(file)

        self.file = file
        output = sys.stdout if output == sys.stdout.name else output

        if isinstance(output, str):
            f = (
                gzip.open(output, "wt")
                if output.endswith(".gz")
                else open(output, "wt")
            )
        else:
            f = output

        fx = self._load_fastx_generator(file)

        for read in fx:
            seq = read[1]

            if seq.upper().__contains__("N"):
                continue
            else:
                f.write("\n".join(read) + "\n")
        f.close()


if __name__ == "__main__":
    # fastx = FastxTools(file='../../data/random_l-180_n-100_hg38_minus-test.fa')
    pass



"""The Cli bioat command entry module

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

>>> example 1:
    in shell:
    $ bioat list

    in python consolo
    >>> from bioat.cli import Cli
    >>> bioat = Cli()
    >>> bioat.list()
    >>> print(bioat.list())
>>> example 2:
    
"""

from bioat import (  # SystemTools,
    BamTools,
    BedTools,
    CrisprTools,
    FastxTools,
    HiCTools,
    MetaTools,
    SearchTools,
    TableTools,
    TargetSeq,
    __version__,
)
from bioat.about import __ABOUT__


class Cli(object):
    """Cli interface of BioAT

    <Brief>
    - "bioat" is short for "Bioinformatic Analysis Tools" and is a commandline toolkit and a python package. It can be used through this cli interface in terminal or the `import` way in python codes.
    - "bioat" has many subcommand to deal with different bio-format: BED, BAM, FASTA, FASTQ, VCF, et al.
    - "bioat" can be used for mining CRISPRs, downloading metagenome and even report a Google Scholar search result!
    - for more information, run `$ bioat about`

    <Copyright>
    For researchers: freely applied to academic research, please cite my work:
            1. #TODO bibtex
            2. #TODO software copyright
    For commercial use:
        NOT PERMITTED unless permission is obtained from the author

    ---

    <COMMAND>
    Such as `bioat version`, it means a directly command

    <GROUPS>
    Such as `bioat crispr`, it is a group/bundle name for commands with a specific topic

    ---

    <Usage>
    A Demo to understand `bioat`:

        bioat about
            <in shell>:
                $ bioat about
            <in python consolo>:
                >>> from bioat.cli import Cli
                >>> bioat = Cli()
                >>> bioat.list()
                >>> print(bioat.list())
    """

    # - Citation:
    #     - Citation is not forcible and I would be appreciated if you wann to
    #  do this.
    #     - citation reference format:
    #         [1] Huanan ZHAO. bioat, a Bioinformatic Analysis Tool kit (2023),
    # https://github.com/hermanzhaozzzz/bioat

    # """

    def __init__(self):
        self.bam = BamTools()
        self.bed = BedTools()
        self.crispr = CrisprTools()
        self.fastx = FastxTools()
        self.hic = HiCTools()
        self.meta = MetaTools()
        self.search = SearchTools()
        # self.system = SystemTools()
        self.table = TableTools()
        self.target_seq = TargetSeq()

    @classmethod
    def about(cls):
        """Return information about `bioat`."""
        return __ABOUT__

    def list(
        self,
    ):
        """Return GROUPS and COMMANDS.

        :return: tree formatted subcommands that can be used
        :rtype: string
        """
        # ! do not change obj method to class method
        # ! because some attr and methods do not exit before obj instantiation
        out = ""
        for att in dir(self):
            if not att.startswith("_"):
                out += f"{att}\n"

                for sub_att in dir(getattr(self, att)):
                    if not sub_att.startswith("_"):
                        out += f"  ├── {sub_att}\n"
        return out

    @classmethod
    def version(cls):
        """Return version information.

        :return: version info
        :rtype: string
        """
        return __version__



"""
BioAt BedTools
# todo
"""

from bioat.lib.libdataclasses import Bed


class BedTools:
    """Bed toolbox."""

    def __init__(self):
        pass

    def random(
        self,
        kind="human",
        length=200,
        number=100,
        fasta=None,
        seed=0,
        rm_n=True,
    ) -> Bed:
        """Get random region

        :param kind: Species:Tag, defaults to "human"
        :type kind: str, optional
        :param length: random region length, defaults to 200
        :type length: int, optional
        :param number: random region number, defaults to 100
        :type number: int, optional
        :param fasta: get random region from which genome, must be defined when rm_n=True, defaults to None
        :type fasta: str, optional
        :param seed: random seed, defaults to 0
        :type seed: int, optional
        :param rm_n: remove random region contains N base (It's often the telomere region), defaults to True
        :type rm_n: bool, optional
        :return: bed file
        :rtype: Bed
        """
        # TODO



import os
import sys

from bioat.logger import get_logger

__module_name__ = "bioat.lib.libsystem"


def check_cmd(x):
    """
    Check if a command is available in system PATH var.
    """
    logger = get_logger(
        level="debug", module_name=__module_name__, func_name="check_cmd"
    )
    logger.info("Checking command '%s'", x)
    return any(
        os.access(os.path.join(path, x), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )



"""bioat.lib.libplot

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

This module provides functions for plotting.

example 1:
    init_matplotlib
        <in python consolo>:
            >>> import matplotlib.pyplot as plt
            >>> from bioat.lib.libplot import init_matplotlib
            >>> init_matplotlib(log_level='info')
            >>> plt.plot([1, 2, 3], [4, 5, 6])
            >>> plt.show()

example 2:
    plot_colortable
        <in python consolo>:
            >>> from bioat.lib.libplot import plot_colortable
            >>> colors = ['#64C1E8', '#80CED7', '#63C7B2', '#8E6C88', '#CA61C3', '#FF958C', '#883677']
            >>> plot_colortable(colors, ncols=1, labels=[1, 2, 3, 4, 5, 6, 7])
            >>> plt.show()
"""

import math
import os
import shutil
import sys
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from dna_features_viewer import GraphicFeature, GraphicRecord
from matplotlib.patches import Rectangle

from bioat.exceptions import BioatException, BioatRuntimeWarning
from bioat.lib.libpath import HOME
from bioat.logger import get_logger

__all__ = ["init_matplotlib", "plot_colortable"]
__module_name__ = "bioat.lib.libplot"

BIOAT_DEFAULT_FONTS_DATAPATH = os.path.join(os.path.dirname(__file__), "libplot")
BIOAT_DEFAULT_FONTS = [
    "Helvetica-Bold.ttf",
    "Helvetica-BoldOblique.ttf",
    "Helvetica-Light.ttf",
    "Helvetica-Oblique.ttf",
    "Helvetica.ttf",
]


def _copy_fonts(refresh=False, log_level="warning"):
    """copy fonts from bioat package to matplotlib package

    :param log_level: log level for logger
    :type log_level: string
    :raises FileNotFoundError: if not found which site-packages to copy fonts to
    """
    logger = get_logger(
        level=log_level,
        module_name=__module_name__,
        func_name="_copy_fonts",
    )
    if refresh:
        files_to_remove = glob(os.path.join(HOME, ".cache", "matplotlib", "fontlist*"))
        for file in files_to_remove:
            os.remove(file)
    try:
        to_path = None
        for i in sys.path:
            if i.endswith("site-packages"):
                to_path = i
                break
        if not to_path:
            logger.warning(BioatRuntimeWarning("site-packages not found in sys.path"))
            return
        to_path = [i for i in sys.path if i.endswith("site-packages")][0]
        to_path = os.path.join(
            to_path, "matplotlib", "mpl-data", "fonts", "ttf"
        )
        from_path = BIOAT_DEFAULT_FONTS_DATAPATH
        logger.debug(f"Copying fonts from {from_path} to {to_path}")
        for font in BIOAT_DEFAULT_FONTS:
            shutil.copyfile(
                os.path.join(from_path, font),
                os.path.join(to_path, font),
            )
        logger.debug("Fonts copied successfully")
    except Exception as e:
        logger.error(BioatException(f"Failed to copy fonts: {e}"))


def init_matplotlib(
    style="ggplot", font="Helvetica", refresh=False, log_level="INFO", **kwargs
):
    """easily set matplotlib style

    :param style: matplotlib style, defaults to 'ggplot'
    :type style: str, optional
    :param font: use what font in matplotlib, defaults to 'Helvetica'
    :type font: str, optional
    :param refresh: wether to remove matplotlib font cache and reset it, defaults to False
    :type refresh: bool, optional
    :param log_level: log level, defaults to 'INFO'
    :type log_level: str, optional
    :param set_backend_pdf: whether to use core fonts for the PDF backend, defaults to True
    :type set_backend_pdf: bool, optional
    :param set_backend_ps: whether to use core fonts for the PS backend, defaults to True
    :type set_backend_ps: bool, optional
    :param set_backend_svg: whether to use 'none' to replace 'path' (use font but not plot path for characters)for the SVG backend, defaults to 'none'
    :type set_backend_svg: str, optional
    :raises BioatException: if failed to copy fonts
    :raises BioatRuntimeWarning: if site-packages not found in sys.path
    """
    logger = get_logger(
        level=log_level,
        module_name=__module_name__,
        func_name="init_matplotlib",
    )
    logger.info('Initializing matplotlib')
    _copy_fonts(refresh=refresh, log_level=log_level)
    logger.info(
        f"set: plt.style.use('{style}')\n"
        "# set matplotlib style theme\n"
        "# ref: https://matplotlib.org/stable/api/style_api.html"
    )
    plt.style.use(style)

    logger.info(f"set: plt.rcParams['font.family'] = '{font}'")
    plt.rcParams["font.family"] = font
    logger.info(f"set: plt.rcParams['font.sans-serif'] = ['{font}']")
    plt.rcParams["font.sans-serif"] = [font]

    # set backends
    set_backend_pdf = kwargs.get("set_backend_pdf", True)
    set_backend_ps = kwargs.get("set_backend_ps", True)
    set_backend_svg = kwargs.get("set_backend_svg", "none")

    if set_backend_pdf:
        logger.info(
            f"set: plt.rcParams['pdf.use14corefonts'] = {set_backend_pdf} \n"
            "# whether to use core fonts for the PDF backend\n"
            "# ref: https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.rcParams)"
        )
        plt.rcParams["pdf.use14corefonts"] = set_backend_pdf
    if set_backend_ps:
        logger.info(
            f"set: plt.rcParams['ps.useafm'] = {set_backend_ps}\n"
            "# whether to use core fonts for the PS backend"
        )
        plt.rcParams["ps.useafm"] = set_backend_ps
    if set_backend_svg:
        logger.info(
            f"set: plt.rcParams['svg.fonttype'] = '{set_backend_svg}'\n"
            "# whether to use 'none' to replace 'path' (use font but not plot path for characters)for the SVG backend"
        )
        plt.rcParams["svg.fonttype"] = set_backend_svg
    logger.info("matplotlib initialized successfully")


def plot_colortable(colors, *, ncols=4):
    cell_width = 212
    cell_height = 22
    swatch_width = 48
    margin = 12

    names = list(colors)

    n = len(names)
    nrows = math.ceil(n / ncols)

    width = cell_width * 4 + 2 * margin
    height = cell_height * nrows + 2 * margin
    dpi = 72

    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin / width, margin / height,
                        (width - margin) / width, (height - margin) / height)
    ax.set_xlim(0, cell_width * 4)
    ax.set_ylim(cell_height * (nrows - 0.5), -cell_height / 2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(
            text_pos_x,
            y,
            names[i],
            fontsize=14,
            horizontalalignment="left",
            verticalalignment="center",
        )

        ax.add_patch(
            Rectangle(xy=(swatch_start_x, y - 9), width=swatch_width,
                      height=18, facecolor=name, edgecolor='0.7')
        )
    plt.show()
    plt.close()


def set_figsize(figsize=(3.5, 2.5)):
    """Set the figure size for matplotlib.

    Defined in :numref:`sec_calculus`"""
    # use_svg_display()
    plt.rcParams["figure.figsize"] = figsize


def set_axes(axes, xlabel, ylabel, xlim, ylim, xscale, yscale, legend):
    """Set the axes for matplotlib.

    Defined in :numref:`sec_calculus`"""
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)
    axes.set_xscale(xscale)
    axes.set_yscale(yscale)
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)

    if legend:
        axes.legend(legend)
    axes.grid()


def plot(
    X,
    Y=None,
    xlabel=None,
    ylabel=None,
    legend=[],
    xlim=None,
    ylim=None,
    xscale="linear",
    yscale="linear",
    fmts=("-", "m--", "g-.", "r:"),
    figsize=(3.5, 2.5),
    axes=None,
):
    """Plot data points.

    Defined in :numref:`sec_calculus`"""

    def has_one_axis(X):  # True if X (tensor or list) has 1 axis
        return (
            hasattr(X, "ndim")
            and X.ndim == 1
            or isinstance(X, list)
            and not hasattr(X[0], "__len__")
        )

    if has_one_axis(X):
        X = [X]
    if Y is None:
        X, Y = [[]] * len(X), X
    elif has_one_axis(Y):
        Y = [Y]
    if len(X) != len(Y):
        X = X * len(Y)

    set_figsize(figsize)
    if axes is None:
        axes = plt.gca()
    axes.cla()
    for x, y, fmt in zip(X, Y, fmts):
        axes.plot(x, y, fmt) if len(x) else axes.plot(y, fmt)
    set_axes(axes, xlabel, ylabel, xlim, ylim, xscale, yscale, legend)
    # return axes


def plot_dna_features(
    df: pd.DataFrame | None = None,
    use_demo_data=True,
    fig_width=10,
    fig_height=2,
    col_locus_start="locus_start",
    col_locus_length="locus_length",
    col_group="group",
    col_name="name",
    col_start="start",
    col_end="end",
    col_strand="strand",
    col_color="color",
    **kwargs,
) -> plt.Figure:
    # """
    # ![](https://raw.githubusercontent.com/hermanzhaozzzz/PicturesBed01/master/PicGo/202406260156158.png)
    # ==================== DEMO ====================
    #     group     name    type  start    end  strand    color  locus_length  locus_start
    # 0    0000  AAAAAAA     CDS   1050   4000     1.0  #ffd700         20000            0
    # 1    0000     BBBB     CDS   3500   6000    -1.0  #ffcccc         20000            0
    # 2    0000       CC     CDS  14000  16000     NaN  #cffccc         20000            0
    # 3    0000     None      DR  10300  10336     NaN    black         20000            0
    # 4    0000     None      DR  10366  10402     NaN    black         20000            0
    # ..    ...      ...     ...    ...    ...     ...      ...           ...          ...
    # 510  4444     None  spacer  12106  12136     NaN      red         14000         1000
    # 511  4444     None  spacer  12172  12202     NaN      red         14000         1000
    # 512  4444     None  spacer  12238  12268     NaN      red         14000         1000
    # 513  4444     None  spacer  12304  12334     NaN      red         14000         1000
    # 514  4444     None  spacer  12370  12400     NaN      red         14000         1000

    # [515 rows x 9 columns]
    # ==================== /DEMO ====================
    # """
    def _get_demo_data():
        """get demo data for DNA features

        :return: demo DNA features
        :rtype: pd.DataFrame
        """
        df = pd.DataFrame(
            {
                "name": ["AAAAAAA", "BBBB", "CC"] + [None] * 50 + [None] * 50,
                "type": ["CDS", "CDS", "CDS"] + ["DR"] * 50 + ["spacer"] * 50,
                "start": (
                    [1050, 3500, 14000]
                    + list(range(10300, 10300 + 66 * 50, 66))
                    + list(range(10336, 10336 + 66 * 50, 66))
                ),
                "end": (
                    [4000, 6000, 16000]
                    + list(range(10336, 10336 + 66 * 50, 66))
                    + list(range(10366, 10366 + 66 * 50, 66))
                ),
                "strand": [1, -1, None] + [None] * 50 + [None] * 50,
                "color": (
                    [
                        "#ffd700",
                        "#ffcccc",
                        "#cffccc",
                    ]
                    + ["black"] * 50
                    + ["red"] * 50
                ),
                "locus_length": [20000] * 103,
                "locus_start": [0] * 103,
            }
        )
        ls = []
        for i in range(5):
            tmpdf = df.copy()
            tmpdf.insert(0, "group", str(i) * 4)
            if i in (2, 4):
                tmpdf["start"] = tmpdf["start"] - 300 * i
                tmpdf["end"] = tmpdf["end"] - 300 * i
                tmpdf["locus_length"] = tmpdf["locus_length"] - 1500 * i
                tmpdf["locus_start"] = 1000
            ls.append(tmpdf)
        df = pd.concat(ls)
        df.reset_index(drop=True, inplace=True)
        print("=" * 20, "DEMO", "=" * 20)
        print(df)
        print("=" * 20, "/DEMO", "=" * 20)
        return df

    if df is None:
        if use_demo_data:
            df = _get_demo_data()
        else:
            raise ValueError("Please provide a DataFrame with DNA features.")

    # 设置子图的数量
    n = df[col_group].nunique()
    # 创建一个n行1列的子图
    fig, axs = plt.subplots(n, 1, figsize=(fig_width, fig_height * n))
    if n == 1:
        # 确保 axs 是一个列表，方便统一处理
        axs = [axs]

    # 循环遍历每个子图并进行绘图
    for i, g in zip(range(n), df.groupby(col_group)):
        group_name, data = g
        # print(f"group_name = {group_name}")

        features = []
        for _, row in data.iterrows():
            label = row.get(col_name, None)
            if isinstance(label, float):
                label = None
            start = row.get(col_start, -100)
            end = row.get(col_end, 0)
            strand = row.get(col_strand, 0)
            color = row.get(col_color, "grey")
            feature = GraphicFeature(
                start=start, end=end, strand=strand, color=color, label=label
            )
            features.append(feature)
        if use_demo_data:
            locus_start = data["locus_start"].values[0]
            locus_length = data["locus_length"].values[0]
        else:
            locus_start = data[col_locus_start].values[0]
            locus_length = data[col_locus_length].values[0]
        record = GraphicRecord(
            first_index=locus_start, sequence_length=locus_length, features=features
        )
        # print(axs)
        # print(n)
        record.plot(  # https://github.com/Edinburgh-Genome-Foundry/DnaFeaturesViewer
            ax=axs[i],
            figure_width=fig_width,
            draw_line=kwargs.get("draw_line", True),
            with_ruler=kwargs.get("with_ruler", True),
            plot_sequence=False,
            annotate_inline=kwargs.get("annotate_inline", True),
            max_label_length=kwargs.get("max_label_length", 50),
            max_line_length=kwargs.get("max_line_length", 20),
            level_offset=0,
            strand_in_label_threshold="default",
            elevate_outline_annotations="default",
            figure_height=fig_height,
            sequence_params=None,
        )
        # ax.set_title(f"Subplot {i+1}")  # 设置子图的标题
        axs[i].set_xlabel(
            group_name,
        )  # 设置x轴标签
        # ax.set_ylabel("Y Label")  # 设置y轴标签
        axs[i].grid(False)  # 显示网格线
        # plt.show()
    plt.tight_layout()
    return fig

    # 调整子图之间的距离


if __name__ == "__main__":
    # test for init_matplotlib
    # init_matplotlib(log_level="warning")
    # plot_colortable(
    #     ["#64C1E8", "#80CED7", "#63C7B2", "#8E6C88", "#CA61C3", "#FF958C", "#883677"],
    #     ncols=1,
    # )

    # def normal(x, mu, sigma):
    #     p = 1 / math.sqrt(2 * math.pi * sigma**2)
    #     return p * np.exp(-0.5 / sigma**2 * (x - mu) ** 2)

    # x = np.arange(-7, 7, 0.01)
    # params = [(0, 1), (0, 2), (3, 1)]
    # plot(
    #     x,
    #     [normal(x, mu, sigma) for mu, sigma in params],
    #     xlabel="x",
    #     ylabel="p(x)",
    #     figsize=(4.5, 2.5),
    #     legend=[f"mean {mu}, std {sigma}" for mu, sigma in params],
    # )
    # plt.show()
    # plt.close()
    df_plot = pd.read_csv("~/demo.tsv", sep="\t")
    # print(df_plot)
    plot_dna_features()
    plot_dna_features(
        df=df_plot,
        # use_demo_data=True
        fig_width=20,
        fig_height=3,
        col_locus_start="locus_start",
        col_locus_length="locus_length",
        col_group="crispr_id:member",
        col_name="pep_name",
        col_start="pep_start",
        col_end="pep_end",
        col_strand="pep_strand",
        col_color="color",
    )

    # fig.savefig("/Users/zhaohuanan/Downloads/test.png", dpi=300)
    # 显示图形
    # plt.show()



"""_summary_

author: Herman Huanan Zhao
email: hermanzhaozzzz@gmail.com
homepage: https://github.com/hermanzhaozzzz

_description_

example 1:
    bioat list
        <in shell>:
            $ bioat list
        <in python consolo>:
            >>> from bioat.cli import Cli
            >>> bioat = Cli()
            >>> bioat.list()
            >>> print(bioat.list())

example 2:
    _example_
"""

import sys

import pandas as pd

from bioat.logger import get_logger

__module_name__ = 'bioat.tabletools'


class TableTools:
    """To integrate tables."""

    def __init__(self):
        pass

    def merge(
            self,
            inputs,
            tags,
            output,
            input_fmt='tsv',
            output_fmt='tsv',
            input_header=False,
            output_header=False,
            log_level='WARNING'
    ):
        """A simple tool to merge same formatted tables from different sample.

        Params
        :param inputs: table files
        :param tags: tags for each table
        :param output: merged file
        :param input_fmt: tsv | csv
        :param output_fmt: tsv | csv
        :param input_header: True | False, input table has header or not
        :param output_header: True | False, output table has header or not
        :param log_level: log status
        """
        logger = get_logger(
            level=log_level, module_name=__module_name__, func_name="merge"
        )
        # fix params
        inputs = list(inputs) if isinstance(inputs, tuple) else inputs.split(',')
        tags = list(tags) if isinstance(tags, tuple) else tags.split(',')

        logger.debug(f'input_fmt={input_fmt}')
        logger.debug(f'output_fmt={output_fmt}')
        logger.debug(f'input_header={input_header}')
        logger.debug(f'output_header={output_header}')

        input_header = 0 if input_header else None
        output = sys.stdout if output == sys.stdout.name else output
        dt_sep = {'csv': ',', 'tsv': '\t'}

        df = pd.DataFrame()

        for file, tag in zip(inputs, tags):
            _df = pd.read_csv(
                file,
                sep=dt_sep[input_fmt],
                header=input_header,
                index_col=None)
            _df.insert(0, column='<sample>', value=tag)
            df = pd.concat([df, _df], axis=0, ignore_index=True)

        df.to_csv(output, sep=dt_sep[output_fmt], header=output_header, index=None)

    def split(
            self,
            input: str,
            n: int,
            output_prefix=None,
            input_fmt='tsv',
            output_fmt='tsv',
            input_header=False,
            output_header=False,
            compress=False,
            log_level='WARNING'
    ):
        """A simple tool to split table into parts.

        Params
        :param input: table to split
        :param n: split table into n parts
        :param output_prefix: name prefix for splitted parts, the same with input if not defined
        :param input_fmt: tsv | csv
        :param output_fmt: tsv | csv
        :param input_header: True | False, input table has header or not
        :param output_header: True | False, output table has header or not
        :param compress: True | False, gzip the output table or not
        :param log_level: log status
        """
        logger = get_logger(
            level=log_level, module_name=__module_name__, func_name="split"
        )
        logger.debug(f'input_fmt={input_fmt}')
        logger.debug(f'output_fmt={output_fmt}')
        logger.debug(f'input_header={input_header}')
        logger.debug(f'output_header={output_header}')
        # fix params
        input_header = 0 if input_header else None
        output_prefix = output_prefix if output_prefix else f'{input}_'
        compress = '.gz' if compress else ''

        dt_sep = {'csv': ',', 'tsv': '\t'}
        # load table
        df = pd.read_csv(input, header=input_header, sep=dt_sep[input_fmt])

        # max lines for each part table: chunk_size
        if df.shape[0] % n == 0:
            chunk_size = df.shape[0] // n
        else:
            chunk_size = df.shape[0] // n + 1

        # write out
        for idx in range(n):
            df_write_out = df.iloc[idx * chunk_size: (idx + 1) * chunk_size, ]
            df_write_out.to_csv(
                f'{output_prefix}{idx}.%s{compress}' % output_fmt,
                index=None,
                header=output_header,
                sep=dt_sep[output_fmt]
            )


if __name__ == "__main__":
    pass



# demo distance_matrix.txt
"""
 , A, B, C, D, E, F
A, 0, 5, 4, 7, 6, 8
B, 5, 0, 7,10, 9,11
C, 4, 7, 0, 7, 6, 8
D, 7,10, 7, 0, 5, 9
E, 6, 9, 6, 5, 0, 8
F, 8,11, 8, 9, 8, 0
"""
# The algorithm currently runs in O( N^3 ) time (empirically, O( N^2.8 ) or so)
# https://github.com/ivo-pontes/NeighborJoiningTree/blob/master/NeighbourJoining.py



import os
import sys
import time

import psutil

from bioat.logger import get_logger

__module_name__ = "bioat.lib._dev_tools"


def _elapsed_since(start):
    return time.strftime("%Mm:%Ss", time.gmtime(time.time() - start))


def _get_process_memory():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss


# def profile(func):
#     """Decorator for calculate runtime of a function."""

#     def wrapper(*args, **kwargs):
#         logger = get_logger(
#             level="INFO", module_name=__module_name__, func_name="profile"
#         )
#         mem_before = _get_process_memory()
#         start = time.time()
#         result = func(*args, **kwargs)
#         elapsed_time = _elapsed_since(start)
#         mem_after = _get_process_memory()
#         logger.info("{}: \n\tmemory before: {:,}, after: {:,}, consumed: {:,}; \n\texec time: {}".format(
#             func.__name__,
#             mem_before, mem_after, mem_after - mem_before,
#             elapsed_time))
#         return result


#     return wrapper
def profile(num_iterations=1):
    """Decorator for calculating the runtime of a function, over multiple iterations.

    :param func: function to be decorated.
    :type func: function
    :param num_iterations: times to iter, defaults to 1
    :type num_iterations: int, optional
    """

    def inner_decorator(func):
        def wrapper(*args, **kwargs):
            logger = get_logger(
                level="INFO", module_name=__name__, func_name=func.__name__
            )

            total_time = 0
            for i in range(num_iterations):
                mem_before = _get_process_memory()
                start_time = time.time()
                result = func(*args, **kwargs)
                elapsed_time = time.time() - start_time
                total_time += elapsed_time
                mem_after = _get_process_memory()

            avg_time = total_time / num_iterations
            text = (
                f"\n[{func.__name__}] performance profiling:\n\t↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓↓\n"
                f"\tIteration num: {num_iterations}x\n"
                f"\tTotal    time: {total_time:.9f}s\n"
                f"\tAverage  time: {avg_time:.9f}s"
            )
            logger.info(text)

            return result

        return wrapper

    return inner_decorator
